from __future__ import annotations

import asyncio
import time
from collections.abc import AsyncIterator, Collection
from contextlib import asynccontextmanager
from datetime import datetime, timedelta, timezone
from types import SimpleNamespace
from typing import Any

import pytest

import app.modules.proxy.load_balancer as load_balancer_module
from app.core.balancer.types import UpstreamError
from app.core.crypto import TokenEncryptor
from app.core.openai.model_registry import ModelRegistrySnapshot
from app.core.utils.time import utcnow
from app.db.models import (
    Account,
    AccountStatus,
    AdditionalUsageHistory,
    StickySession,
    StickySessionKind,
    UsageHistory,
)
from app.modules.accounts.repository import AccountsRepository
from app.modules.api_keys.repository import ApiKeysRepository
from app.modules.proxy.load_balancer import (
    ADDITIONAL_QUOTA_DATA_UNAVAILABLE,
    NO_ADDITIONAL_QUOTA_ELIGIBLE_ACCOUNTS,
    NO_PLAN_SUPPORT_FOR_MODEL,
    LoadBalancer,
    RuntimeState,
)
from app.modules.proxy.repo_bundle import ProxyRepositories
from app.modules.proxy.sticky_repository import StickySessionsRepository
from app.modules.request_logs.repository import RequestLogsRepository
from app.modules.upstream_identities.repository import OpenAIPlatformIdentitiesRepository
from app.modules.usage.repository import AdditionalUsageRepository, UsageRepository

pytestmark = pytest.mark.unit


def _make_account(account_id: str, email: str = "a@example.com") -> Account:
    encryptor = TokenEncryptor()
    return Account(
        id=account_id,
        chatgpt_account_id=f"workspace-{account_id}",
        email=email,
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access"),
        refresh_token_encrypted=encryptor.encrypt("refresh"),
        id_token_encrypted=encryptor.encrypt("id"),
        last_refresh=datetime.now(tz=timezone.utc),
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )


class StubAccountsRepository(AccountsRepository):
    def __init__(self, accounts: list[Account]) -> None:
        self._accounts = accounts
        self.status_updates: list[dict[str, Any]] = []

    async def get_by_id(self, account_id: str) -> Account | None:
        return self._find_account(account_id)

    async def list_accounts(self) -> list[Account]:
        return list(self._accounts)

    def _find_account(self, account_id: str) -> Account | None:
        return next((account for account in self._accounts if account.id == account_id), None)

    async def update_status(
        self,
        account_id: str,
        status: AccountStatus,
        deactivation_reason: str | None = None,
        reset_at: int | None = None,
    ) -> bool:
        account = self._find_account(account_id)
        if account is None:
            return False
        account.status = status
        account.deactivation_reason = deactivation_reason
        account.reset_at = reset_at
        self.status_updates.append(
            {
                "account_id": account_id,
                "status": status,
                "deactivation_reason": deactivation_reason,
                "reset_at": reset_at,
            }
        )
        return True

    async def update_status_if_current(
        self,
        account_id: str,
        status: AccountStatus,
        deactivation_reason: str | None = None,
        reset_at: int | None = None,
        *,
        expected_status: AccountStatus,
        expected_deactivation_reason: str | None = None,
        expected_reset_at: int | None = None,
    ) -> bool:
        account = self._find_account(account_id)
        if account is None:
            return False
        if (
            account.status != expected_status
            or account.deactivation_reason != expected_deactivation_reason
            or account.reset_at != expected_reset_at
        ):
            return False
        return await self.update_status(account_id, status, deactivation_reason, reset_at)


class StubUsageRepository(UsageRepository):
    def __init__(
        self,
        primary: dict[str, UsageHistory],
        secondary: dict[str, UsageHistory],
    ) -> None:
        self._primary = primary
        self._secondary = secondary
        self.primary_calls = 0
        self.secondary_calls = 0

    async def latest_by_account(self, window: str | None = None) -> dict[str, UsageHistory]:
        if window == "secondary":
            self.secondary_calls += 1
            return self._secondary
        self.primary_calls += 1
        return self._primary


class StubStickySessionsRepository(StickySessionsRepository):
    def __init__(self) -> None:
        pass

    async def get_account_id(
        self,
        key: str,
        *,
        kind: StickySessionKind,
        max_age_seconds: int | None = None,
    ) -> str | None:
        return None

    async def upsert(self, key: str, account_id: str, *, kind: StickySessionKind) -> StickySession:
        return self._build_row(key, account_id, kind)

    async def delete(self, key: str, *, kind: StickySessionKind | None = None) -> bool:
        return False

    @staticmethod
    def _build_row(key: str, account_id: str, kind: StickySessionKind) -> StickySession:
        return StickySession(
            key=key,
            kind=kind,
            provider_kind="chatgpt_web",
            routing_subject_id=account_id,
            account_id=account_id,
        )


class StubRequestLogsRepository(RequestLogsRepository):
    def __init__(self) -> None:
        pass


class StubApiKeysRepository(ApiKeysRepository):
    def __init__(self) -> None:
        pass


class StubPlatformIdentitiesRepository(OpenAIPlatformIdentitiesRepository):
    def __init__(self) -> None:
        pass


class StubAdditionalUsageRepository(AdditionalUsageRepository):
    def __init__(
        self,
        primary: dict[str, AdditionalUsageHistory] | None = None,
        secondary: dict[str, AdditionalUsageHistory] | None = None,
    ) -> None:
        self._primary = primary or {}
        self._secondary = secondary or {}

    async def latest_by_account(
        self,
        quota_key: str | None = None,
        window: str | None = None,
        *,
        limit_name: str | None = None,
        account_ids: Collection[str] | None = None,
        since: datetime | None = None,
    ) -> dict[str, AdditionalUsageHistory]:
        effective_key = quota_key or limit_name
        assert effective_key is not None
        assert window is not None
        if window == "secondary":
            source = self._secondary
        else:
            source = self._primary
        rows = {
            account_id: entry
            for account_id, entry in source.items()
            if getattr(entry, "quota_key", entry.limit_name) == effective_key
        }
        if account_ids is not None:
            account_id_set = set(account_ids)
            rows = {account_id: entry for account_id, entry in rows.items() if account_id in account_id_set}
        if since is not None:
            rows = {account_id: entry for account_id, entry in rows.items() if entry.recorded_at >= since}
        return dict(rows)

    async def latest_by_quota_key(
        self,
        quota_key: str,
        window: str,
        *,
        account_ids: Collection[str] | None = None,
        since: datetime | None = None,
    ) -> dict[str, AdditionalUsageHistory]:
        return await self.latest_by_account(
            quota_key=quota_key,
            window=window,
            account_ids=account_ids,
            since=since,
        )


def _additional_entry(
    entry_id: int,
    *,
    account_id: str,
    window: str,
    used_percent: float,
    recorded_at: datetime | None = None,
    limit_name: str = "GPT-5.3-Codex-Spark",
    quota_key: str = "codex_spark",
    reset_at: int = 1741500000,
) -> AdditionalUsageHistory:
    now = recorded_at or utcnow()
    return AdditionalUsageHistory(
        id=entry_id,
        account_id=account_id,
        quota_key=quota_key,
        limit_name=limit_name,
        metered_feature="codex_bengalfox",
        window=window,
        used_percent=used_percent,
        reset_at=reset_at,
        window_minutes=5 if window == "primary" else 10080,
        recorded_at=now,
    )


@asynccontextmanager
async def _repo_factory(
    accounts_repo: StubAccountsRepository,
    usage_repo: StubUsageRepository,
    sticky_repo: StubStickySessionsRepository,
    additional_usage_repo: StubAdditionalUsageRepository | None = None,
) -> AsyncIterator[ProxyRepositories]:
    yield ProxyRepositories(
        accounts=accounts_repo,
        platform_identities=StubPlatformIdentitiesRepository(),
        usage=usage_repo,
        request_logs=StubRequestLogsRepository(),
        sticky_sessions=sticky_repo,
        api_keys=StubApiKeysRepository(),
        additional_usage=additional_usage_repo or StubAdditionalUsageRepository(),
    )


@pytest.mark.asyncio
async def test_select_account_reads_cached_usage_once_per_window() -> None:
    account = _make_account("acc-load-balancer")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()

    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    selection = await balancer.select_account()

    assert selection.account is not None
    assert selection.account.id == account.id
    assert usage_repo.primary_calls == 1
    assert usage_repo.secondary_calls == 1


@pytest.mark.asyncio
async def test_select_account_filters_to_assigned_account_ids() -> None:
    preferred = _make_account("acc-preferred", "preferred@example.com")
    assigned = _make_account("acc-assigned", "assigned@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    primary = {
        preferred.id: UsageHistory(
            id=1,
            account_id=preferred.id,
            recorded_at=now,
            window="primary",
            used_percent=1.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
        ),
        assigned.id: UsageHistory(
            id=2,
            account_id=assigned.id,
            recorded_at=now,
            window="primary",
            used_percent=90.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
        ),
    }
    secondary = {
        preferred.id: UsageHistory(
            id=3,
            account_id=preferred.id,
            recorded_at=now,
            window="secondary",
            used_percent=1.0,
            reset_at=now_epoch + 3600,
            window_minutes=60,
        ),
        assigned.id: UsageHistory(
            id=4,
            account_id=assigned.id,
            recorded_at=now,
            window="secondary",
            used_percent=90.0,
            reset_at=now_epoch + 3600,
            window_minutes=60,
        ),
    }

    accounts_repo = StubAccountsRepository([preferred, assigned])
    usage_repo = StubUsageRepository(primary=primary, secondary=secondary)
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))

    selection = await balancer.select_account(account_ids=[assigned.id])

    assert selection.account is not None
    assert selection.account.id == assigned.id


@pytest.mark.asyncio
async def test_select_account_scope_does_not_prune_runtime_for_other_accounts() -> None:
    retained = _make_account("acc-retained", "retained@example.com")
    assigned = _make_account("acc-assigned", "assigned@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    primary = {
        retained.id: UsageHistory(
            id=1,
            account_id=retained.id,
            recorded_at=now,
            window="primary",
            used_percent=10.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
        ),
        assigned.id: UsageHistory(
            id=2,
            account_id=assigned.id,
            recorded_at=now,
            window="primary",
            used_percent=20.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
        ),
    }
    secondary = {
        retained.id: UsageHistory(
            id=3,
            account_id=retained.id,
            recorded_at=now,
            window="secondary",
            used_percent=10.0,
            reset_at=now_epoch + 3600,
            window_minutes=60,
        ),
        assigned.id: UsageHistory(
            id=4,
            account_id=assigned.id,
            recorded_at=now,
            window="secondary",
            used_percent=20.0,
            reset_at=now_epoch + 3600,
            window_minutes=60,
        ),
    }

    accounts_repo = StubAccountsRepository([retained, assigned])
    usage_repo = StubUsageRepository(primary=primary, secondary=secondary)
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    balancer._runtime[retained.id] = RuntimeState(cooldown_until=time.time() + 300.0, error_count=2)

    selection = await balancer.select_account(account_ids=[assigned.id])

    assert selection.account is not None
    assert selection.account.id == assigned.id
    assert retained.id in balancer._runtime
    assert balancer._runtime[retained.id].cooldown_until is not None
    assert balancer._runtime[retained.id].error_count == 2


@pytest.mark.asyncio
async def test_select_account_empty_explicit_scope_fails_closed() -> None:
    preferred = _make_account("acc-preferred", "preferred@example.com")
    fallback = _make_account("acc-fallback", "fallback@example.com")
    accounts_repo = StubAccountsRepository([preferred, fallback])
    usage_repo = StubUsageRepository(primary={}, secondary={})
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))

    selection = await balancer.select_account(account_ids=[])

    assert selection.account is None


@pytest.mark.asyncio
async def test_select_account_uses_cached_usage_without_inline_refresh(monkeypatch) -> None:
    async def fail_refresh_accounts(
        self,
        accounts: list[Account],
        latest_usage: dict[str, UsageHistory],
    ) -> bool:
        raise AssertionError("select_account should not refresh usage inline")

    monkeypatch.setattr(
        "app.modules.usage.updater.UsageUpdater.refresh_accounts",
        fail_refresh_accounts,
    )

    account = _make_account("acc-cached-selection")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=15.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()

    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    selection = await balancer.select_account()

    assert selection.account is not None
    assert selection.account.id == account.id
    assert usage_repo.primary_calls == 1
    assert usage_repo.secondary_calls == 1


@pytest.mark.asyncio
async def test_select_account_proceeds_without_cached_usage_rows(monkeypatch) -> None:
    async def fail_refresh_accounts(
        self,
        accounts: list[Account],
        latest_usage: dict[str, UsageHistory],
    ) -> bool:
        raise AssertionError("select_account should not refresh usage inline")

    monkeypatch.setattr(
        "app.modules.usage.updater.UsageUpdater.refresh_accounts",
        fail_refresh_accounts,
    )

    account = _make_account("acc-no-usage-yet")
    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={}, secondary={})
    sticky_repo = StubStickySessionsRepository()

    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    selection = await balancer.select_account()

    assert selection.account is not None
    assert selection.account.id == account.id
    assert usage_repo.primary_calls == 1
    assert usage_repo.secondary_calls == 1


@pytest.mark.asyncio
async def test_select_account_prefilters_accounts_by_additional_usage_limit() -> None:
    account_ineligible = _make_account("acc-additional-exhausted", email="full@example.com")
    account_eligible = _make_account("acc-additional-eligible", email="ok@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_entry = UsageHistory(
        id=1,
        account_id=account_ineligible.id,
        recorded_at=now,
        window="primary",
        used_percent=20.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    primary_entry_ok = UsageHistory(
        id=2,
        account_id=account_eligible.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )

    accounts_repo = StubAccountsRepository([account_ineligible, account_eligible])
    usage_repo = StubUsageRepository(
        primary={account_ineligible.id: primary_entry, account_eligible.id: primary_entry_ok},
        secondary={},
    )
    sticky_repo = StubStickySessionsRepository()
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            account_ineligible.id: _additional_entry(
                11,
                account_id=account_ineligible.id,
                window="primary",
                used_percent=100.0,
                recorded_at=now,
            ),
            account_eligible.id: _additional_entry(
                12,
                account_id=account_eligible.id,
                window="primary",
                used_percent=35.0,
                recorded_at=now,
            ),
        }
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            accounts_repo,
            usage_repo,
            sticky_repo,
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(
        additional_limit_name="codex_spark",
        routing_strategy="usage_weighted",
    )

    assert selection.account is not None
    assert selection.account.id == account_eligible.id


@pytest.mark.asyncio
async def test_select_account_requires_fresh_additional_usage_data(monkeypatch) -> None:
    monkeypatch.setattr(
        "app.core.config.settings.get_settings",
        lambda: SimpleNamespace(usage_refresh_interval_seconds=600),
    )

    account_stale = _make_account("acc-additional-stale", email="stale@example.com")
    account_fresh = _make_account("acc-additional-fresh", email="fresh@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    usage_rows = {
        account_stale.id: UsageHistory(
            id=21,
            account_id=account_stale.id,
            recorded_at=now,
            window="primary",
            used_percent=15.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
        ),
        account_fresh.id: UsageHistory(
            id=22,
            account_id=account_fresh.id,
            recorded_at=now,
            window="primary",
            used_percent=10.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
        ),
    }
    accounts_repo = StubAccountsRepository([account_stale, account_fresh])
    usage_repo = StubUsageRepository(primary=usage_rows, secondary={})
    sticky_repo = StubStickySessionsRepository()
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            account_stale.id: _additional_entry(
                31,
                account_id=account_stale.id,
                window="primary",
                used_percent=5.0,
                recorded_at=now - timedelta(seconds=1201),
            ),
            account_fresh.id: _additional_entry(
                32,
                account_id=account_fresh.id,
                window="primary",
                used_percent=5.0,
                recorded_at=now - timedelta(seconds=1199),
            ),
        }
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            accounts_repo,
            usage_repo,
            sticky_repo,
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(additional_limit_name="codex_spark")

    assert selection.account is not None
    assert selection.account.id == account_fresh.id


@pytest.mark.asyncio
async def test_select_account_uses_canonical_quota_key_for_upstream_limit_alias(monkeypatch) -> None:
    account = _make_account("acc-additional-alias", email="alias@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    usage_repo = StubUsageRepository(
        primary={
            account.id: UsageHistory(
                id=41,
                account_id=account.id,
                recorded_at=now,
                window="primary",
                used_percent=10.0,
                reset_at=now_epoch + 300,
                window_minutes=5,
            )
        },
        secondary={},
    )
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            account.id: _additional_entry(
                42,
                account_id=account.id,
                window="primary",
                limit_name="GPT-5.3-Codex-Spark",
                quota_key="codex_spark",
                used_percent=5.0,
                recorded_at=now,
            )
        }
    )

    monkeypatch.setattr(
        "app.modules.proxy.load_balancer.get_model_registry",
        lambda: SimpleNamespace(plan_types_for_model=lambda _model: frozenset({"plus"})),
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            StubAccountsRepository([account]),
            usage_repo,
            StubStickySessionsRepository(),
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(model="gpt-5.3-codex-spark")

    assert selection.account is not None
    assert selection.account.id == account.id


@pytest.mark.asyncio
@pytest.mark.parametrize("additional_limit_name", ["codex_other", "GPT-5.3-Codex-Spark"])
async def test_select_account_accepts_legacy_additional_limit_aliases(additional_limit_name: str) -> None:
    account = _make_account(f"acc-additional-{additional_limit_name}")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    usage_repo = StubUsageRepository(
        primary={
            account.id: UsageHistory(
                id=51,
                account_id=account.id,
                recorded_at=now,
                window="primary",
                used_percent=10.0,
                reset_at=now_epoch + 300,
                window_minutes=5,
            )
        },
        secondary={},
    )
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            account.id: _additional_entry(
                52,
                account_id=account.id,
                window="primary",
                limit_name="GPT-5.3-Codex-Spark",
                quota_key="codex_spark",
                used_percent=5.0,
                recorded_at=now,
            )
        }
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            StubAccountsRepository([account]),
            usage_repo,
            StubStickySessionsRepository(),
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(additional_limit_name=additional_limit_name)

    assert selection.account is not None
    assert selection.account.id == account.id


@pytest.mark.asyncio
async def test_select_account_prunes_stale_runtime_for_removed_accounts() -> None:
    account_id = "acc-reused"
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account(account_id)
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([])
    usage_repo = StubUsageRepository(primary={}, secondary={account_id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    balancer._runtime[account_id] = RuntimeState(cooldown_until=time.time() + 300.0)

    empty_selection = await balancer.select_account()
    assert empty_selection.account is None
    assert account_id not in balancer._runtime

    accounts_repo._accounts = [account]
    usage_repo._primary = {account_id: primary_entry}

    selection = await balancer.select_account()
    assert selection.account is not None
    assert selection.account.id == account_id


@pytest.mark.asyncio
async def test_round_robin_does_not_serialize_concurrent_selection(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account_a = _make_account("acc-round-robin-a", "a@example.com")
    account_b = _make_account("acc-round-robin-b", "b@example.com")
    primary_entries = {
        account_a.id: UsageHistory(
            id=1,
            account_id=account_a.id,
            recorded_at=now,
            window="primary",
            used_percent=10.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
        ),
        account_b.id: UsageHistory(
            id=2,
            account_id=account_b.id,
            recorded_at=now,
            window="primary",
            used_percent=10.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
        ),
    }
    secondary_entries = {
        account_a.id: UsageHistory(
            id=3,
            account_id=account_a.id,
            recorded_at=now,
            window="secondary",
            used_percent=10.0,
            reset_at=now_epoch + 3600,
            window_minutes=60,
        ),
        account_b.id: UsageHistory(
            id=4,
            account_id=account_b.id,
            recorded_at=now,
            window="secondary",
            used_percent=10.0,
            reset_at=now_epoch + 3600,
            window_minutes=60,
        ),
    }

    accounts_repo = StubAccountsRepository([account_a, account_b])
    usage_repo = StubUsageRepository(primary=primary_entries, secondary=secondary_entries)
    sticky_repo = StubStickySessionsRepository()

    original_persist_selection_state = LoadBalancer._persist_selection_state
    overlap_observed = asyncio.Event()
    inflight_persist_calls = 0

    async def slow_persist_selection_state(
        self: LoadBalancer,
        accounts_repo: AccountsRepository,
        account_map: dict[str, Account],
        states: list[Any],
    ) -> None:
        nonlocal inflight_persist_calls
        inflight_persist_calls += 1
        try:
            if inflight_persist_calls >= 2:
                overlap_observed.set()
            await asyncio.sleep(0.05)
            await original_persist_selection_state(self, accounts_repo, account_map, states)
        finally:
            inflight_persist_calls -= 1

    monkeypatch.setattr(LoadBalancer, "_persist_selection_state", slow_persist_selection_state)

    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    start = asyncio.Event()

    async def pick_account() -> str:
        await start.wait()
        selection = await balancer.select_account(routing_strategy="round_robin")
        assert selection.account is not None
        return selection.account.id

    first = asyncio.create_task(pick_account())
    second = asyncio.create_task(pick_account())
    started = time.perf_counter()
    start.set()
    selected_ids = await asyncio.gather(first, second)
    elapsed = time.perf_counter() - started

    assert len(set(selected_ids)) == 2
    assert overlap_observed.is_set()
    assert elapsed < 0.13


@pytest.mark.asyncio
async def test_select_account_does_not_clobber_concurrent_error_state(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-runtime-race", "race@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()

    original_persist_selection_state = LoadBalancer._persist_selection_state
    release_select_sync = asyncio.Event()
    select_sync_blocked = asyncio.Event()
    blocked_once = False

    async def controlled_persist_selection_state(
        self: LoadBalancer,
        accounts_repo: AccountsRepository,
        account_map: dict[str, Account],
        states: list[Any],
    ) -> None:
        nonlocal blocked_once
        if not blocked_once and any(state.error_count == 0 for state in states):
            blocked_once = True
            select_sync_blocked.set()
            await release_select_sync.wait()
        await original_persist_selection_state(self, accounts_repo, account_map, states)

    monkeypatch.setattr(LoadBalancer, "_persist_selection_state", controlled_persist_selection_state)

    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    select_task = asyncio.create_task(balancer.select_account())
    await select_sync_blocked.wait()

    record_error_task = asyncio.create_task(balancer.record_error(account))
    await asyncio.sleep(0.01)
    assert record_error_task.done()

    release_select_sync.set()
    await select_task
    await record_error_task

    runtime = balancer._runtime[account.id]
    assert runtime.error_count == 1
    assert runtime.last_error_at is not None


@pytest.mark.asyncio
async def test_mark_quota_exceeded_keeps_selection_blocked_until_persisted(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-quota-lock", "quota-lock@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    persist_started = asyncio.Event()
    release_persist = asyncio.Event()

    async def blocking_update_status(
        account_id: str,
        status: AccountStatus,
        deactivation_reason: str | None = None,
        reset_at: int | None = None,
    ) -> bool:
        persist_started.set()
        await release_persist.wait()
        return await StubAccountsRepository.update_status(
            accounts_repo,
            account_id,
            status,
            deactivation_reason,
            reset_at,
        )

    monkeypatch.setattr(accounts_repo, "update_status", blocking_update_status)

    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    quota_error: UpstreamError = {"message": "quota exceeded"}
    mark_task = asyncio.create_task(balancer.mark_quota_exceeded(account, quota_error))
    await persist_started.wait()

    select_task = asyncio.create_task(balancer.select_account())
    await asyncio.sleep(0.01)
    assert not select_task.done()

    release_persist.set()
    await mark_task
    await select_task

    assert accounts_repo.status_updates[0]["status"] == AccountStatus.QUOTA_EXCEEDED


@pytest.mark.asyncio
async def test_record_errors_does_not_restore_terminal_status(monkeypatch) -> None:
    account = _make_account("acc-record-errors-race", "record-errors-race@example.com")
    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={}, secondary={})
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))

    original_persist_state_if_current = balancer._persist_state_if_current
    persist_started = asyncio.Event()
    release_persist = asyncio.Event()

    async def blocking_persist_state_if_current(
        accounts_repo_arg: AccountsRepository,
        account_arg: Account,
        state_arg: Any,
    ) -> bool:
        persist_started.set()
        await release_persist.wait()
        return await original_persist_state_if_current(accounts_repo_arg, account_arg, state_arg)

    monkeypatch.setattr(balancer, "_persist_state_if_current", blocking_persist_state_if_current)

    record_task = asyncio.create_task(balancer.record_errors(account, 1))
    await persist_started.wait()

    fail_task = asyncio.create_task(balancer.mark_permanent_failure(account, "refresh_token_expired"))
    await asyncio.sleep(0.01)
    assert not fail_task.done()

    release_persist.set()
    await record_task
    await fail_task

    assert account.status == AccountStatus.DEACTIVATED
    assert accounts_repo.status_updates[-1]["status"] == AccountStatus.DEACTIVATED
    assert all(update["status"] != AccountStatus.ACTIVE for update in accounts_repo.status_updates)


@pytest.mark.asyncio
async def test_select_account_does_not_hold_runtime_lock_during_input_loading(monkeypatch) -> None:
    accounts_started = asyncio.Event()
    release_accounts = asyncio.Event()

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-refresh-unblocks-runtime", "runtime@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()

    async def blocking_list_accounts() -> list[Account]:
        accounts_started.set()
        await release_accounts.wait()
        return [account]

    monkeypatch.setattr(accounts_repo, "list_accounts", blocking_list_accounts)

    @asynccontextmanager
    async def repo_factory() -> AsyncIterator[ProxyRepositories]:
        yield ProxyRepositories(
            accounts=accounts_repo,
            platform_identities=StubPlatformIdentitiesRepository(),
            usage=usage_repo,
            additional_usage=StubAdditionalUsageRepository(),
            request_logs=object(),  # type: ignore[arg-type]
            sticky_sessions=sticky_repo,
            api_keys=object(),  # type: ignore[arg-type]
        )

    balancer = LoadBalancer(repo_factory)
    select_task = asyncio.create_task(balancer.select_account())
    await accounts_started.wait()

    record_error_task = asyncio.create_task(balancer.record_error(account))
    await asyncio.sleep(0.01)

    assert record_error_task.done()
    runtime = balancer._runtime[account.id]
    assert runtime.error_count == 1
    assert runtime.last_error_at is not None

    release_accounts.set()
    selection = await select_task
    assert selection.account is not None


@pytest.mark.asyncio
async def test_select_account_does_not_open_repo_before_runtime_lock(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-runtime-before-repo", "runtime-before-repo@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=20.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    repo_entered = asyncio.Event()
    release_repo = asyncio.Event()

    @asynccontextmanager
    async def blocking_repo_factory() -> AsyncIterator[ProxyRepositories]:
        repo_entered.set()
        await release_repo.wait()
        yield ProxyRepositories(
            accounts=accounts_repo,
            platform_identities=StubPlatformIdentitiesRepository(),
            usage=usage_repo,
            additional_usage=StubAdditionalUsageRepository(),
            request_logs=StubRequestLogsRepository(),
            sticky_sessions=sticky_repo,
            api_keys=StubApiKeysRepository(),
        )

    balancer = LoadBalancer(blocking_repo_factory)

    async def fake_load_selection_inputs(
        *,
        model: str | None,
        additional_limit_name: str | None = None,
        account_ids: Collection[str] | None = None,
    ):
        del model, additional_limit_name, account_ids
        return load_balancer_module._SelectionInputs(
            accounts=[account],
            latest_primary={account.id: primary_entry},
            latest_secondary={account.id: secondary_entry},
        )

    monkeypatch.setattr(balancer, "_load_selection_inputs", fake_load_selection_inputs)

    # T21 made select_account lock-free (per-account locking replaces global _runtime_lock).
    # select_account now proceeds without acquiring _runtime_lock.
    # Verify that select_account still works correctly without the global lock.
    release_repo.set()
    selection = await balancer.select_account()
    assert repo_entered.is_set()
    assert selection.account is not None


@pytest.mark.asyncio
async def test_select_account_skips_stale_persistence_after_terminal_status_update(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-select-lockstep", "select-lockstep@example.com")
    account.status = AccountStatus.QUOTA_EXCEEDED
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=20.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    persist_blocked = asyncio.Event()
    release_persist = asyncio.Event()
    original_update_status_if_current = accounts_repo.update_status_if_current

    async def blocking_update_status_if_current(
        account_id: str,
        status: AccountStatus,
        deactivation_reason: str | None = None,
        reset_at: int | None = None,
        *,
        expected_status: AccountStatus,
        expected_deactivation_reason: str | None = None,
        expected_reset_at: int | None = None,
    ) -> bool:
        persist_blocked.set()
        await release_persist.wait()
        return await original_update_status_if_current(
            account_id,
            status,
            deactivation_reason,
            reset_at,
            expected_status=expected_status,
            expected_deactivation_reason=expected_deactivation_reason,
            expected_reset_at=expected_reset_at,
        )

    monkeypatch.setattr(accounts_repo, "update_status_if_current", blocking_update_status_if_current)

    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    select_task = asyncio.create_task(balancer.select_account())
    await persist_blocked.wait()

    fail_task = asyncio.create_task(balancer.mark_permanent_failure(account, "refresh_token_expired"))
    await fail_task

    release_persist.set()
    selection = await select_task

    assert accounts_repo.status_updates[-1]["status"] == AccountStatus.DEACTIVATED
    assert selection.account is None


@pytest.mark.asyncio
async def test_select_account_retries_after_post_persist_permanent_failure(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-post-persist-deactivate", "post-persist-deactivate@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))

    original_persist_selection_state = balancer._persist_selection_state
    injected = False

    async def wrapped_persist_selection_state(accounts_repo_arg, account_map, states):
        nonlocal injected
        result = await original_persist_selection_state(accounts_repo_arg, account_map, states)
        if not injected:
            injected = True
            await balancer.mark_permanent_failure(account, "refresh_token_expired")
        return result

    monkeypatch.setattr(balancer, "_persist_selection_state", wrapped_persist_selection_state)

    selection = await balancer.select_account()

    assert account.status == AccountStatus.DEACTIVATED
    assert selection.account is None


@pytest.mark.asyncio
async def test_select_account_retries_after_post_persist_quota_exceeded(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-post-persist-quota", "post-persist-quota@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))

    original_persist_selection_state = balancer._persist_selection_state
    injected = False

    async def wrapped_persist_selection_state(accounts_repo_arg, account_map, states):
        nonlocal injected
        result = await original_persist_selection_state(accounts_repo_arg, account_map, states)
        if not injected:
            injected = True
            await balancer.mark_quota_exceeded(account, {"message": "quota exceeded"})
        return result

    monkeypatch.setattr(balancer, "_persist_selection_state", wrapped_persist_selection_state)

    selection = await balancer.select_account()

    assert account.status == AccountStatus.QUOTA_EXCEEDED
    assert selection.account is None


@pytest.mark.asyncio
async def test_sync_runtime_state_bumps_version_for_status_only_updates() -> None:
    account = _make_account("acc-status-only-version", "status-only-version@example.com")
    balancer = LoadBalancer(
        lambda: _repo_factory(
            StubAccountsRepository([]),
            StubUsageRepository({}, {}),
            StubStickySessionsRepository(),
        )
    )
    runtime = balancer._runtime.setdefault(account.id, RuntimeState())
    initial_version = runtime.version

    state = load_balancer_module.AccountState(
        account_id=account.id,
        status=AccountStatus.DEACTIVATED,
        deactivation_reason="Refresh token expired - re-login required",
    )

    updated = balancer._sync_runtime_state(account, state)

    assert updated is True
    assert balancer._runtime[account.id].version == initial_version + 1


@pytest.mark.skip(reason="T21 per-account locking eliminates version conflicts that this test was designed to catch")
@pytest.mark.asyncio
async def test_select_account_reloads_inputs_after_version_conflict(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-reload-after-conflict", "reload-after-conflict@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))

    original_load_selection_inputs = balancer._load_selection_inputs
    load_calls = 0

    async def counted_load_selection_inputs(
        *,
        model: str | None,
        additional_limit_name: str | None = None,
        account_ids: Collection[str] | None = None,
    ):
        nonlocal load_calls
        load_calls += 1
        return await original_load_selection_inputs(
            model=model,
            additional_limit_name=additional_limit_name,
            account_ids=account_ids,
        )

    original_select_account = load_balancer_module.select_account
    first_call = True

    def conflict_injecting_select_account(states, **kwargs):
        nonlocal first_call
        if first_call:
            first_call = False
            account.status = AccountStatus.DEACTIVATED
            account.deactivation_reason = "Refresh token expired - re-login required"
            balancer._runtime.setdefault(account.id, RuntimeState()).version += 1
        return original_select_account(states, **kwargs)

    monkeypatch.setattr(balancer, "_load_selection_inputs", counted_load_selection_inputs)
    monkeypatch.setattr(load_balancer_module, "select_account", conflict_injecting_select_account)

    selection = await balancer.select_account()

    assert load_calls >= 2
    assert selection.account is None


@pytest.mark.skip(reason="T21 per-account locking eliminates version conflicts that this test was designed to catch")
@pytest.mark.asyncio
async def test_select_account_does_not_hold_runtime_lock_during_conflict_reload(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-conflict-reload-unblocks-runtime", "conflict-reload@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))

    original_load_selection_inputs = balancer._load_selection_inputs
    reload_started = asyncio.Event()
    release_reload = asyncio.Event()
    load_calls = 0

    async def blocking_load_selection_inputs(*, model: str | None, additional_limit_name: str | None = None):
        nonlocal load_calls
        load_calls += 1
        if load_calls == 2:
            reload_started.set()
            await release_reload.wait()
        return await original_load_selection_inputs(model=model, additional_limit_name=additional_limit_name)

    original_select_account = load_balancer_module.select_account
    first_call = True

    def conflict_injecting_select_account(states, **kwargs):
        nonlocal first_call
        if first_call:
            first_call = False
            balancer._runtime.setdefault(account.id, RuntimeState()).version += 1
        return original_select_account(states, **kwargs)

    monkeypatch.setattr(balancer, "_load_selection_inputs", blocking_load_selection_inputs)
    monkeypatch.setattr(load_balancer_module, "select_account", conflict_injecting_select_account)

    select_task = asyncio.create_task(balancer.select_account())
    await reload_started.wait()

    record_error_task = asyncio.create_task(balancer.record_error(account))
    await asyncio.sleep(0.01)

    assert record_error_task.done()
    runtime = balancer._runtime[account.id]
    assert runtime.error_count == 1
    assert runtime.last_error_at is not None

    release_reload.set()
    selection = await select_task
    assert selection.account is not None


@pytest.mark.skip(reason="T21 per-account locking eliminates version conflicts that this test was designed to catch")
@pytest.mark.asyncio
async def test_select_account_sticky_reloads_inputs_after_stale_selected_persistence(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-sticky-stale-selected", "sticky-stale-selected@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))

    original_load_selection_inputs = balancer._load_selection_inputs
    load_calls = 0

    async def counted_load_selection_inputs(
        *,
        model: str | None,
        additional_limit_name: str | None = None,
        account_ids: Collection[str] | None = None,
    ):
        nonlocal load_calls
        load_calls += 1
        return await original_load_selection_inputs(
            model=model,
            additional_limit_name=additional_limit_name,
            account_ids=account_ids,
        )

    async def pinned_account_id(
        key: str,
        *,
        kind: StickySessionKind,
        max_age_seconds: int | None = None,
    ) -> str | None:
        del key, kind, max_age_seconds
        return account.id

    original_persist_selection_state = balancer._persist_selection_state
    first_persist = True

    async def stale_selected_persist(
        accounts_repo: AccountsRepository,
        account_map: dict[str, Account],
        states: list[Any],
    ) -> set[str]:
        nonlocal first_persist
        if first_persist:
            first_persist = False
            account.status = AccountStatus.DEACTIVATED
            account.deactivation_reason = "Refresh token expired - re-login required"
            return {account.id}
        return await original_persist_selection_state(accounts_repo, account_map, states)

    monkeypatch.setattr(balancer, "_load_selection_inputs", counted_load_selection_inputs)
    monkeypatch.setattr(sticky_repo, "get_account_id", pinned_account_id)
    monkeypatch.setattr(balancer, "_persist_selection_state", stale_selected_persist)

    selection = await balancer.select_account(
        sticky_key="sticky-session-1",
        sticky_kind=StickySessionKind.CODEX_SESSION,
    )

    assert load_calls >= 2
    assert selection.account is None


@pytest.mark.asyncio
async def test_select_account_sticky_does_not_return_stale_selection_at_retry_cap(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-sticky-stale-retry-cap", "sticky-stale-retry-cap@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))

    original_load_selection_inputs = balancer._load_selection_inputs
    load_calls = 0

    async def counted_load_selection_inputs(
        *,
        model: str | None,
        additional_limit_name: str | None = None,
        account_ids: Collection[str] | None = None,
    ):
        nonlocal load_calls
        load_calls += 1
        return await original_load_selection_inputs(
            model=model,
            additional_limit_name=additional_limit_name,
            account_ids=account_ids,
        )

    async def pinned_account_id(
        key: str,
        *,
        kind: StickySessionKind,
        max_age_seconds: int | None = None,
    ) -> str | None:
        del key, kind, max_age_seconds
        return account.id

    async def always_stale_selected_persist(
        accounts_repo: AccountsRepository,
        account_map: dict[str, Account],
        states: list[Any],
    ) -> set[str]:
        del accounts_repo, account_map, states
        return {account.id}

    monkeypatch.setattr(balancer, "_load_selection_inputs", counted_load_selection_inputs)
    monkeypatch.setattr(sticky_repo, "get_account_id", pinned_account_id)
    monkeypatch.setattr(balancer, "_persist_selection_state", always_stale_selected_persist)

    selection = await balancer.select_account(
        sticky_key="sticky-session-retry-cap",
        sticky_kind=StickySessionKind.CODEX_SESSION,
    )

    assert load_calls >= 2
    assert selection.account is None


@pytest.mark.asyncio
async def test_select_account_skips_registry_plan_filter_for_mapped_model(monkeypatch) -> None:
    account = _make_account("acc-gated-registry-skip", "gated-registry-skip@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=5.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={})
    sticky_repo = StubStickySessionsRepository()
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            account.id: _additional_entry(
                2,
                account_id=account.id,
                window="primary",
                used_percent=20.0,
                reset_at=now_epoch + 300,
                recorded_at=now,
            )
        }
    )

    monkeypatch.setattr(
        "app.modules.proxy.load_balancer.get_model_registry",
        lambda: SimpleNamespace(
            get_snapshot=lambda: SimpleNamespace(model_plans={}),
            plan_types_for_model=lambda _model: frozenset(),
        ),
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            accounts_repo,
            usage_repo,
            sticky_repo,
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(model="gpt-5.3-codex-spark")

    assert selection.account is not None
    assert selection.account.id == account.id
    assert selection.error_code is None


@pytest.mark.asyncio
async def test_select_account_respects_registry_plan_filter_for_mapped_model(monkeypatch) -> None:
    account = _make_account("acc-gated-plan-filtered", "gated-plan-filtered@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=5.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={})
    sticky_repo = StubStickySessionsRepository()
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            account.id: _additional_entry(
                2,
                account_id=account.id,
                window="primary",
                used_percent=20.0,
                reset_at=now_epoch + 300,
                recorded_at=now,
            )
        }
    )

    monkeypatch.setattr(
        "app.modules.proxy.load_balancer.get_model_registry",
        lambda: SimpleNamespace(
            get_snapshot=lambda: ModelRegistrySnapshot(
                models={},
                model_plans={"gpt-5.3-codex-spark": frozenset({"pro"})},
                plan_models={"pro": frozenset({"gpt-5.3-codex-spark"})},
                fetched_at=0.0,
            ),
            plan_types_for_model=lambda _model: frozenset({"pro"}),
        ),
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            accounts_repo,
            usage_repo,
            sticky_repo,
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(model="gpt-5.3-codex-spark")

    assert selection.account is None
    assert selection.error_code == NO_PLAN_SUPPORT_FOR_MODEL


@pytest.mark.asyncio
async def test_select_account_returns_plan_support_error_for_ungated_model(monkeypatch) -> None:
    account = _make_account("acc-ungated-plan-filtered", "ungated-plan-filtered@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=5.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={})
    sticky_repo = StubStickySessionsRepository()

    monkeypatch.setattr(
        "app.modules.proxy.load_balancer.get_model_registry",
        lambda: SimpleNamespace(plan_types_for_model=lambda _model: frozenset({"pro"})),
    )

    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    selection = await balancer.select_account(model="gpt-5.3-codex")

    assert selection.account is None
    assert selection.error_code == NO_PLAN_SUPPORT_FOR_MODEL
    assert selection.error_message == "No accounts with a plan supporting model 'gpt-5.3-codex'"


@pytest.mark.asyncio
async def test_select_account_empty_pool_preserves_no_accounts_for_modeled_request(monkeypatch) -> None:
    accounts_repo = StubAccountsRepository([])
    usage_repo = StubUsageRepository(primary={}, secondary={})
    sticky_repo = StubStickySessionsRepository()

    monkeypatch.setattr(
        "app.modules.proxy.load_balancer.get_model_registry",
        lambda: SimpleNamespace(plan_types_for_model=lambda _model: frozenset({"pro"})),
    )

    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    selection = await balancer.select_account(model="gpt-5.3-codex")

    assert selection.account is None
    assert selection.error_code is None
    assert selection.error_message is not None
    assert "No available accounts" in selection.error_message


@pytest.mark.asyncio
async def test_select_account_retries_no_accounts_after_runtime_recovery(monkeypatch) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = _make_account("acc-no-accounts-retry", "no-accounts-retry@example.com")
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=10.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary_entry = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=10.0,
        reset_at=now_epoch + 3600,
        window_minutes=60,
    )

    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={account.id: secondary_entry})
    sticky_repo = StubStickySessionsRepository()
    balancer = LoadBalancer(lambda: _repo_factory(accounts_repo, usage_repo, sticky_repo))
    balancer._runtime[account.id] = RuntimeState(error_count=3, last_error_at=time.time())

    original_persist_selection_state = balancer._persist_selection_state
    persist_started = asyncio.Event()
    release_persist = asyncio.Event()

    async def blocking_persist_selection_state(
        accounts_repo_arg: AccountsRepository,
        account_map: dict[str, Account],
        states: list[Any],
    ) -> set[str]:
        persist_started.set()
        await release_persist.wait()
        return await original_persist_selection_state(accounts_repo_arg, account_map, states)

    monkeypatch.setattr(balancer, "_persist_selection_state", blocking_persist_selection_state)

    select_task = asyncio.create_task(balancer.select_account())
    await persist_started.wait()

    await balancer.record_success(account)
    release_persist.set()
    selection = await select_task

    assert selection.account is not None
    assert selection.account.id == account.id


@pytest.mark.asyncio
async def test_select_account_returns_data_unavailable_error_for_mapped_model(monkeypatch) -> None:
    account = _make_account("acc-gated-stale", "gated-stale@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=5.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={})
    sticky_repo = StubStickySessionsRepository()
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            account.id: _additional_entry(
                2,
                account_id=account.id,
                window="primary",
                used_percent=20.0,
                recorded_at=now - timedelta(seconds=181),
            )
        }
    )

    monkeypatch.setattr(
        "app.modules.proxy.load_balancer.get_model_registry",
        lambda: SimpleNamespace(plan_types_for_model=lambda _model: frozenset({"plus"})),
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            accounts_repo,
            usage_repo,
            sticky_repo,
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(model="gpt-5.3-codex-spark")

    assert selection.account is None
    assert selection.error_code == ADDITIONAL_QUOTA_DATA_UNAVAILABLE


@pytest.mark.asyncio
async def test_select_account_returns_data_unavailable_when_secondary_window_is_stale(monkeypatch) -> None:
    account = _make_account("acc-gated-stale-secondary", "gated-stale-secondary@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=5.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={})
    sticky_repo = StubStickySessionsRepository()
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            account.id: _additional_entry(
                2,
                account_id=account.id,
                window="primary",
                used_percent=20.0,
                reset_at=now_epoch + 300,
                recorded_at=now,
            )
        },
        secondary={
            account.id: _additional_entry(
                3,
                account_id=account.id,
                window="secondary",
                used_percent=20.0,
                reset_at=now_epoch + 3600,
                recorded_at=now - timedelta(seconds=181),
            )
        },
    )

    monkeypatch.setattr(
        "app.modules.proxy.load_balancer.get_model_registry",
        lambda: SimpleNamespace(plan_types_for_model=lambda _model: frozenset({"plus"})),
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            accounts_repo,
            usage_repo,
            sticky_repo,
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(model="gpt-5.3-codex-spark")

    assert selection.account is None
    assert selection.error_code == ADDITIONAL_QUOTA_DATA_UNAVAILABLE


@pytest.mark.asyncio
async def test_select_account_allows_primary_only_account_when_other_account_has_secondary_history(
    monkeypatch,
) -> None:
    primary_only_account = _make_account("acc-primary-only", "primary-only@example.com")
    stale_secondary_account = _make_account("acc-stale-secondary", "stale-secondary@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    usage_rows = {
        primary_only_account.id: UsageHistory(
            id=1,
            account_id=primary_only_account.id,
            recorded_at=now,
            window="primary",
            used_percent=5.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
        ),
        stale_secondary_account.id: UsageHistory(
            id=2,
            account_id=stale_secondary_account.id,
            recorded_at=now,
            window="primary",
            used_percent=5.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
        ),
    }
    accounts_repo = StubAccountsRepository([primary_only_account, stale_secondary_account])
    usage_repo = StubUsageRepository(primary=usage_rows, secondary={})
    sticky_repo = StubStickySessionsRepository()
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            primary_only_account.id: _additional_entry(
                11,
                account_id=primary_only_account.id,
                window="primary",
                used_percent=20.0,
                reset_at=now_epoch + 300,
                recorded_at=now,
            ),
            stale_secondary_account.id: _additional_entry(
                12,
                account_id=stale_secondary_account.id,
                window="primary",
                used_percent=20.0,
                reset_at=now_epoch + 300,
                recorded_at=now,
            ),
        },
        secondary={
            stale_secondary_account.id: _additional_entry(
                13,
                account_id=stale_secondary_account.id,
                window="secondary",
                used_percent=20.0,
                reset_at=now_epoch + 3600,
                recorded_at=now - timedelta(seconds=181),
            ),
        },
    )

    monkeypatch.setattr(
        "app.modules.proxy.load_balancer.get_model_registry",
        lambda: SimpleNamespace(plan_types_for_model=lambda _model: frozenset({"plus"})),
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            accounts_repo,
            usage_repo,
            sticky_repo,
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(model="gpt-5.3-codex-spark")

    assert selection.account is not None
    assert selection.account.id == primary_only_account.id
    assert selection.error_code is None


@pytest.mark.asyncio
async def test_select_account_returns_no_eligible_error_for_mapped_model(monkeypatch) -> None:
    account = _make_account("acc-gated-exhausted", "gated-exhausted@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=5.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={})
    sticky_repo = StubStickySessionsRepository()
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            account.id: _additional_entry(
                2,
                account_id=account.id,
                window="primary",
                used_percent=100.0,
                reset_at=now_epoch + 300,
                recorded_at=now,
            )
        },
        secondary={
            account.id: _additional_entry(
                3,
                account_id=account.id,
                window="secondary",
                used_percent=10.0,
                reset_at=now_epoch + 3600,
                recorded_at=now,
            )
        },
    )

    monkeypatch.setattr(
        "app.modules.proxy.load_balancer.get_model_registry",
        lambda: SimpleNamespace(plan_types_for_model=lambda _model: frozenset({"plus"})),
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            accounts_repo,
            usage_repo,
            sticky_repo,
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(model="gpt-5.3-codex-spark")

    assert selection.account is None
    assert selection.error_code == NO_ADDITIONAL_QUOTA_ELIGIBLE_ACCOUNTS


@pytest.mark.asyncio
async def test_select_account_additional_limit_filter_does_not_mutate_account_status(monkeypatch) -> None:
    account = _make_account("acc-gated-status-stable", "status-stable@example.com")
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_entry = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=5.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    accounts_repo = StubAccountsRepository([account])
    usage_repo = StubUsageRepository(primary={account.id: primary_entry}, secondary={})
    sticky_repo = StubStickySessionsRepository()
    additional_usage_repo = StubAdditionalUsageRepository(
        primary={
            account.id: _additional_entry(
                2,
                account_id=account.id,
                window="primary",
                used_percent=20.0,
                reset_at=now_epoch + 300,
                recorded_at=now,
            )
        }
    )

    monkeypatch.setattr(
        "app.modules.proxy.load_balancer.get_model_registry",
        lambda: SimpleNamespace(plan_types_for_model=lambda _model: frozenset({"plus"})),
    )

    balancer = LoadBalancer(
        lambda: _repo_factory(
            accounts_repo,
            usage_repo,
            sticky_repo,
            additional_usage_repo,
        )
    )
    selection = await balancer.select_account(model="gpt-5.3-codex-spark")

    assert selection.account is not None
    assert selection.account.id == account.id
    assert accounts_repo.status_updates == []
    assert account.status == AccountStatus.ACTIVE
    assert account.deactivation_reason is None



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timezone

from app.core import usage as usage_core
from app.core.usage.logs import cached_input_tokens_from_log, cost_from_log, total_tokens_from_log
from app.core.usage.types import (
    BucketModelAggregate,
    RequestActivityAggregate,
    UsageCostByModel,
    UsageCostSummary,
    UsageMetricsSummary,
    UsageSummaryPayload,
    UsageWindowRow,
    UsageWindowSnapshot,
)
from app.core.utils.time import from_epoch_seconds
from app.db.models import Account, AdditionalUsageHistory, RequestLog
from app.modules.usage.schemas import (
    MetricsTrends,
    TrendPoint,
    UsageCost,
    UsageHistoryItem,
    UsageHistoryResponse,
    UsageMetrics,
    UsageSummaryResponse,
    UsageWindow,
    UsageWindowResponse,
)

_BUCKET_COUNT = 28
_BUCKET_SECONDS = 21600  # 6 hours


@dataclass(frozen=True)
class ActivityCostSummary:
    currency: str
    total_usd: float
    by_model: list[UsageCostByModel]


@dataclass(frozen=True)
class ActivityMetricsSummary:
    requests: int | None
    tokens: int | None
    cached_input_tokens: int | None = None
    error_rate: float | None = None
    error_count: int | None = None
    top_error: str | None = None


def align_bucket_window_start(
    since: datetime,
    bucket_seconds: int,
) -> datetime:
    since_epoch = (
        int(since.replace(tzinfo=timezone.utc).timestamp()) if since.tzinfo is None else int(since.timestamp())
    )
    remainder = since_epoch % bucket_seconds
    first_bucket_epoch = since_epoch if remainder == 0 else since_epoch - remainder + bucket_seconds
    aligned = datetime.fromtimestamp(first_bucket_epoch, tz=timezone.utc)
    if since.tzinfo is None:
        return aligned.replace(tzinfo=None)
    return aligned


def build_trends_from_buckets(
    rows: list[BucketModelAggregate],
    since: datetime,
    bucket_seconds: int = _BUCKET_SECONDS,
    bucket_count: int = _BUCKET_COUNT,
    top_error: str | None = None,
) -> tuple[MetricsTrends, ActivityMetricsSummary, ActivityCostSummary]:
    # Align slots so the last slot contains "now" (since + window).
    # Use floor to snap since to a bucket boundary, then shift by 1
    # so that recent data falls within the slot range.
    first_bucket = align_bucket_window_start(since, bucket_seconds)
    first_bucket_epoch = (
        int(first_bucket.replace(tzinfo=timezone.utc).timestamp())
        if first_bucket.tzinfo is None
        else int(first_bucket.timestamp())
    )
    slots = [first_bucket_epoch + i * bucket_seconds for i in range(bucket_count)]
    slot_set = set(slots)

    # Accumulate per-bucket values
    bucket_requests: dict[int, int] = defaultdict(int)
    bucket_errors: dict[int, int] = defaultdict(int)
    bucket_tokens: dict[int, int] = defaultdict(int)
    bucket_costs: dict[int, float] = defaultdict(float)
    total_costs_by_model: dict[str, float] = defaultdict(float)

    total_requests = 0
    total_errors = 0
    total_tokens = 0
    total_cached_tokens = 0
    total_cost_usd = 0.0

    for row in rows:
        epoch = row.bucket_epoch
        if epoch not in slot_set:
            continue
        bucket_requests[epoch] += row.request_count
        bucket_errors[epoch] += row.error_count
        bucket_tokens[epoch] += row.input_tokens + row.output_tokens
        bucket_costs[epoch] += float(row.cost_usd)
        total_costs_by_model[row.model] += float(row.cost_usd)

        total_requests += row.request_count
        total_errors += row.error_count
        total_tokens += row.input_tokens + row.output_tokens
        total_cached_tokens += row.cached_input_tokens
        total_cost_usd += float(row.cost_usd)

    requests_points: list[TrendPoint] = []
    tokens_points: list[TrendPoint] = []
    cost_points: list[TrendPoint] = []
    error_rate_points: list[TrendPoint] = []

    for epoch in slots:
        t = datetime.fromtimestamp(epoch, tz=timezone.utc)
        req = bucket_requests.get(epoch, 0)
        err = bucket_errors.get(epoch, 0)
        tok = bucket_tokens.get(epoch, 0)
        cost_value = bucket_costs.get(epoch, 0.0)

        err_rate = (err / req) if req > 0 else 0.0

        requests_points.append(TrendPoint(t=t, v=float(req)))
        tokens_points.append(TrendPoint(t=t, v=float(tok)))
        cost_points.append(TrendPoint(t=t, v=round(cost_value, 6)))
        error_rate_points.append(TrendPoint(t=t, v=round(err_rate, 4)))

    trends = MetricsTrends(
        requests=requests_points,
        tokens=tokens_points,
        cost=cost_points,
        error_rate=error_rate_points,
    )

    error_rate_total: float | None = None
    if total_requests > 0:
        error_rate_total = total_errors / total_requests

    metrics = ActivityMetricsSummary(
        requests=total_requests,
        tokens=total_tokens,
        cached_input_tokens=total_cached_tokens,
        error_rate=error_rate_total,
        error_count=total_errors,
        top_error=top_error,
    )

    total_cost = ActivityCostSummary(
        currency="USD",
        total_usd=round(total_cost_usd, 6),
        by_model=[
            UsageCostByModel(model=model, usd=round(cost, 6)) for model, cost in sorted(total_costs_by_model.items())
        ],
    )

    return trends, metrics, total_cost


def build_activity_summaries(
    aggregate: RequestActivityAggregate,
    *,
    top_error: str | None = None,
) -> tuple[ActivityMetricsSummary, ActivityCostSummary]:
    total_requests = aggregate.request_count
    total_tokens = aggregate.input_tokens + aggregate.output_tokens
    error_rate: float | None = None
    if total_requests > 0:
        error_rate = aggregate.error_count / total_requests

    return (
        ActivityMetricsSummary(
            requests=total_requests,
            tokens=total_tokens,
            cached_input_tokens=aggregate.cached_input_tokens,
            error_rate=error_rate,
            error_count=aggregate.error_count,
            top_error=top_error,
        ),
        ActivityCostSummary(
            currency="USD",
            total_usd=round(aggregate.cost_usd, 6),
            by_model=[],
        ),
    )


def build_usage_summary_response(
    *,
    accounts: list[Account],
    primary_rows: list[UsageWindowRow],
    secondary_rows: list[UsageWindowRow],
    logs_secondary: list[RequestLog],
    metrics_override: UsageMetricsSummary | None = None,
    cost_override: UsageCostSummary | None = None,
) -> UsageSummaryResponse:
    account_map = {account.id: account for account in accounts}
    primary_window = usage_core.summarize_usage_window(primary_rows, account_map, "primary")
    secondary_window = usage_core.summarize_usage_window(secondary_rows, account_map, "secondary")

    if cost_override is not None:
        cost = cost_override
    else:
        cost = _cost_summary_from_logs(logs_secondary)

    metrics = metrics_override if metrics_override is not None else _usage_metrics(logs_secondary)

    payload = usage_core.parse_usage_summary(primary_window, secondary_window, cost, metrics)
    return _summary_payload_to_response(payload)


def build_usage_history_response(
    *,
    hours: int,
    usage_rows: list[UsageWindowRow],
    accounts: list[Account],
    window: str,
) -> UsageHistoryResponse:
    account_map = {account.id: account for account in accounts}
    accounts_history = _build_account_history(
        usage_rows,
        account_map,
        window,
        missing_remaining_percent=100.0,
    )
    return UsageHistoryResponse(window_hours=hours, accounts=accounts_history)


def build_usage_window_response(
    *,
    window_key: str,
    window_minutes: int | None,
    usage_rows: list[UsageWindowRow],
    accounts: list[Account],
) -> UsageWindowResponse:
    account_map = {account.id: account for account in accounts}
    accounts_history = _build_account_history(
        usage_rows,
        account_map,
        window_key,
        missing_remaining_percent=None,
    )
    return UsageWindowResponse(
        window_key=window_key,
        window_minutes=window_minutes,
        accounts=accounts_history,
    )


def _build_account_history(
    usage_rows: list[UsageWindowRow],
    account_map: dict[str, Account],
    window: str,
    *,
    missing_remaining_percent: float | None,
) -> list[UsageHistoryItem]:
    usage_by_account = {row.account_id: row for row in usage_rows}

    results: list[UsageHistoryItem] = []
    for account_id, account in account_map.items():
        usage = usage_by_account.get(account_id)
        used_percent = usage.used_percent if usage else None
        used_percent_value = float(used_percent) if used_percent is not None else None
        remaining_percent = usage_core.remaining_percent_from_used(used_percent_value)
        if remaining_percent is None:
            remaining_percent = missing_remaining_percent
        capacity = usage_core.capacity_for_plan(account.plan_type, window)
        remaining_credits = usage_core.remaining_credits_from_percent(used_percent_value, capacity)
        if remaining_credits is None and missing_remaining_percent is not None:
            remaining_credits = capacity
        results.append(
            UsageHistoryItem(
                account_id=account_id,
                remaining_percent_avg=remaining_percent,
                capacity_credits=float(capacity or 0.0),
                remaining_credits=float(remaining_credits or 0.0),
            )
        )
    return results


def _cost_summary_from_logs(logs: list[RequestLog]) -> UsageCostSummary:
    total = 0.0
    by_model: dict[str, float] = defaultdict(float)
    for log in logs:
        cost = cost_from_log(log)
        if cost is None:
            continue
        total += cost
        by_model[log.model] += cost
    return UsageCostSummary(
        currency="USD",
        total_usd_7d=round(total, 6),
        by_model=[UsageCostByModel(model=model, usd=round(cost, 6)) for model, cost in sorted(by_model.items())],
    )


def _usage_metrics(logs_secondary: list[RequestLog]) -> UsageMetricsSummary:
    total_requests = len(logs_secondary)
    error_logs = [log for log in logs_secondary if log.status != "success"]
    error_rate: float | None = None
    if total_requests > 0:
        error_rate = len(error_logs) / total_requests
    top_error = _top_error_code(error_logs)
    tokens_secondary = _sum_tokens(logs_secondary)
    cached_tokens_secondary = _sum_cached_input_tokens(logs_secondary)
    return UsageMetricsSummary(
        requests_7d=total_requests,
        tokens_secondary_window=tokens_secondary,
        cached_tokens_secondary_window=cached_tokens_secondary,
        error_rate_7d=error_rate,
        top_error=top_error,
    )


def _sum_tokens(logs: list[RequestLog]) -> int:
    total = 0
    for log in logs:
        total += total_tokens_from_log(log) or 0
    return total


def _sum_cached_input_tokens(logs: list[RequestLog]) -> int:
    total = 0
    for log in logs:
        total += cached_input_tokens_from_log(log) or 0
    return total


def _top_error_code(logs: list[RequestLog]) -> str | None:
    counts: dict[str, int] = {}
    for log in logs:
        code = log.error_code
        if not code:
            continue
        counts[code] = counts.get(code, 0) + 1
    if not counts:
        return None
    return max(counts.items(), key=lambda item: item[1])[0]


def _summary_payload_to_response(payload: UsageSummaryPayload) -> UsageSummaryResponse:
    return UsageSummaryResponse(
        primary_window=build_usage_window_summary_model(payload.primary_window),
        secondary_window=(
            build_usage_window_summary_model(payload.secondary_window) if payload.secondary_window else None
        ),
        cost=_cost_summary_to_model(payload.cost),
        metrics=_metrics_summary_to_model(payload.metrics) if payload.metrics else None,
    )


def build_usage_window_summary_model(snapshot: UsageWindowSnapshot) -> UsageWindow:
    capacity_credits = float(snapshot.capacity_credits)
    remaining_credits = usage_core.remaining_credits_from_used(snapshot.used_credits, capacity_credits) or 0.0
    remaining_percent = max(0.0, 100.0 - float(snapshot.used_percent)) if capacity_credits > 0 else 0.0
    return UsageWindow(
        remaining_percent=remaining_percent,
        capacity_credits=capacity_credits,
        remaining_credits=remaining_credits,
        reset_at=from_epoch_seconds(snapshot.reset_at),
        window_minutes=snapshot.window_minutes,
    )


def _cost_summary_to_model(cost: UsageCostSummary) -> UsageCost:
    return UsageCost(
        currency=cost.currency,
        totalUsd7d=cost.total_usd_7d,
    )


def _metrics_summary_to_model(metrics: UsageMetricsSummary) -> UsageMetrics:
    return UsageMetrics(
        requests_7d=metrics.requests_7d,
        tokens_secondary_window=metrics.tokens_secondary_window,
        cached_tokens_secondary_window=metrics.cached_tokens_secondary_window,
        error_rate_7d=metrics.error_rate_7d,
        top_error=metrics.top_error,
    )


# ---------------------------------------------------------------------------
# Additional usage aggregation
# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class AdditionalWindowSummary:
    """Summary for one window (primary or secondary) of an additional rate limit."""

    used_percent: float  # average across accounts
    reset_at: int | None  # max reset_at across accounts
    window_minutes: int | None  # max window_minutes


@dataclass(frozen=True)
class AdditionalQuotaSummary:
    """Aggregated summary for one additional rate limit (e.g., codex_other)."""

    limit_name: str
    metered_feature: str
    primary_window: AdditionalWindowSummary | None
    secondary_window: AdditionalWindowSummary | None


def build_additional_usage_summary(
    additional_usage_data: dict[str, dict[str, dict[str, AdditionalUsageHistory]]],
) -> list[AdditionalQuotaSummary]:
    """Build aggregated additional quota summaries from per-account data.

    Args:
        additional_usage_data: Nested mapping of
            ``limit_name -> window_key -> account_id -> AdditionalUsageHistory``.

    Returns:
        One :class:`AdditionalQuotaSummary` per *limit_name* present in the input.
    """
    results: list[AdditionalQuotaSummary] = []

    for limit_name, windows in additional_usage_data.items():
        primary_entries = windows.get("primary", {})
        secondary_entries = windows.get("secondary", {})

        # Derive metered_feature from any available entry
        metered_feature = _metered_feature_from_entries(primary_entries, secondary_entries)

        primary_window = _aggregate_additional_window(primary_entries)
        secondary_window = _aggregate_additional_window(secondary_entries)

        results.append(
            AdditionalQuotaSummary(
                limit_name=limit_name,
                metered_feature=metered_feature,
                primary_window=primary_window,
                secondary_window=secondary_window,
            )
        )

    return results


def _aggregate_additional_window(
    entries: dict[str, AdditionalUsageHistory],
) -> AdditionalWindowSummary | None:
    """Aggregate per-account entries into a single window summary.

    Averaging ``used_percent``, using the earliest ``reset_at`` (min) and the
    largest ``window_minutes`` (max) for consistent pool behavior.
    """
    if not entries:
        return None

    total_percent = 0.0
    count = 0
    reset_candidates: list[int] = []
    wm_candidates: list[int] = []

    for entry in entries.values():
        total_percent += entry.used_percent
        count += 1
        if entry.reset_at is not None:
            reset_candidates.append(entry.reset_at)
        if entry.window_minutes is not None:
            wm_candidates.append(entry.window_minutes)

    return AdditionalWindowSummary(
        used_percent=total_percent / count,
        reset_at=min(reset_candidates) if reset_candidates else None,
        window_minutes=max(wm_candidates) if wm_candidates else None,
    )


def _metered_feature_from_entries(
    primary: dict[str, AdditionalUsageHistory],
    secondary: dict[str, AdditionalUsageHistory],
) -> str:
    """Extract ``metered_feature`` from the first available entry."""
    for entries in (primary, secondary):
        for entry in entries.values():
            return entry.metered_feature
    return ""



"""add api_keys table and request_logs.api_key_id

Revision ID: 20260213_000800_add_api_keys
Revises: 20260213_000700_add_dashboard_settings_password
Create Date: 2026-02-13
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260213_000800_add_api_keys"
down_revision = "20260213_000700_add_dashboard_settings_password"
branch_labels = None
depends_on = None


def _table_exists(connection: Connection, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    return inspector.has_table(table_name)


def _columns(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {column["name"] for column in inspector.get_columns(table_name)}


def _indexes(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {str(index["name"]) for index in inspector.get_indexes(table_name) if index.get("name") is not None}


def upgrade() -> None:
    bind = op.get_bind()

    if not _table_exists(bind, "api_keys"):
        op.create_table(
            "api_keys",
            sa.Column("id", sa.String(), primary_key=True),
            sa.Column("name", sa.String(), nullable=False),
            sa.Column("key_hash", sa.String(), nullable=False, unique=True),
            sa.Column("key_prefix", sa.String(), nullable=False),
            sa.Column("allowed_models", sa.Text(), nullable=True),
            sa.Column("weekly_token_limit", sa.Integer(), nullable=True),
            sa.Column("weekly_tokens_used", sa.Integer(), nullable=False, server_default=sa.text("0")),
            sa.Column("weekly_reset_at", sa.DateTime(), nullable=False),
            sa.Column("expires_at", sa.DateTime(), nullable=True),
            sa.Column("is_active", sa.Boolean(), nullable=False, server_default=sa.true()),
            sa.Column(
                "created_at",
                sa.DateTime(),
                nullable=False,
                server_default=sa.text("CURRENT_TIMESTAMP"),
            ),
            sa.Column("last_used_at", sa.DateTime(), nullable=True),
        )

    existing_indexes = _indexes(bind, "api_keys")
    if "idx_api_keys_hash" not in existing_indexes:
        op.create_index("idx_api_keys_hash", "api_keys", ["key_hash"], unique=False)

    request_logs_columns = _columns(bind, "request_logs")
    if request_logs_columns and "api_key_id" not in request_logs_columns:
        with op.batch_alter_table("request_logs") as batch_op:
            batch_op.add_column(sa.Column("api_key_id", sa.String(), nullable=True))


def downgrade() -> None:
    bind = op.get_bind()

    request_logs_columns = _columns(bind, "request_logs")
    if "api_key_id" in request_logs_columns:
        with op.batch_alter_table("request_logs") as batch_op:
            batch_op.drop_column("api_key_id")

    if _table_exists(bind, "api_keys"):
        existing_indexes = _indexes(bind, "api_keys")
        if "idx_api_keys_hash" in existing_indexes:
            op.drop_index("idx_api_keys_hash", table_name="api_keys")
        op.drop_table("api_keys")



from __future__ import annotations

import inspect
import logging
import time
from collections.abc import AsyncIterator, Awaitable, Callable, Mapping, Sequence
from dataclasses import dataclass, replace
from typing import Protocol, cast

from app.core.clients.openai_platform import OpenAIPlatformError, validate_platform_identity
from app.core.clients.openai_platform import create_response as create_platform_response
from app.core.clients.openai_platform import fetch_models as fetch_platform_models
from app.core.clients.openai_platform import stream_responses as stream_platform_responses
from app.core.clients.proxy import compact_responses as _proxy_compact_responses
from app.core.clients.proxy import stream_responses as _proxy_stream_responses
from app.core.clients.proxy import transcribe_audio as _proxy_transcribe_audio
from app.core.clients.proxy_websocket import UpstreamResponsesWebSocket, connect_responses_websocket
from app.core.config.settings import get_settings
from app.core.crypto import TokenEncryptor
from app.core.openai.models import CompactResponsePayload, OpenAIResponsePayload
from app.core.openai.requests import ResponsesCompactRequest, ResponsesRequest
from app.core.types import JsonValue
from app.core.utils.json_guards import is_json_mapping
from app.core.utils.request_id import get_request_id
from app.db.models import Account
from app.modules.accounts.auth_manager import AuthManager
from app.modules.proxy.helpers import _header_account_id
from app.modules.proxy.repo_bundle import ProxyRepoFactory, ProxyRepositories
from app.modules.upstream_identities.types import (
    CHATGPT_WEB_PROVIDER_KIND,
    OPENAI_PLATFORM_PROVIDER_KIND,
    OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
    PHASE1_PLATFORM_ROUTE_FAMILIES,
    ProviderKind,
)
from app.modules.usage.updater import UsageUpdater

logger = logging.getLogger(__name__)

type _CompactResponsesCallable = Callable[
    [ResponsesCompactRequest, Mapping[str, str], str, str | None],
    Awaitable[CompactResponsePayload],
]
type _TranscribeAudioCallable = Callable[
    ...,
    Awaitable[dict[str, JsonValue]],
]


async def core_compact_responses(
    payload: ResponsesCompactRequest,
    headers: Mapping[str, str],
    access_token: str,
    account_id: str | None,
) -> CompactResponsePayload:
    return await _proxy_compact_responses(payload, headers, access_token, account_id)


def core_stream_responses(
    payload: ResponsesRequest,
    headers: Mapping[str, str],
    access_token: str,
    account_id: str | None,
    *,
    raise_for_status: bool = True,
    upstream_stream_transport_override: str | None = None,
) -> AsyncIterator[str]:
    return _proxy_stream_responses(
        payload,
        headers,
        access_token,
        account_id,
        raise_for_status=raise_for_status,
        upstream_stream_transport_override=upstream_stream_transport_override,
    )


async def core_transcribe_audio(
    audio_bytes: bytes,
    *,
    filename: str,
    content_type: str | None,
    prompt: str | None,
    headers: Mapping[str, str],
    access_token: str,
    account_id: str | None,
) -> dict[str, JsonValue]:
    return await _proxy_transcribe_audio(
        audio_bytes,
        filename=filename,
        content_type=content_type,
        prompt=prompt,
        headers=headers,
        access_token=access_token,
        account_id=account_id,
    )


_DEFAULT_CORE_COMPACT_RESPONSES = core_compact_responses
_DEFAULT_CORE_STREAM_RESPONSES = core_stream_responses
_DEFAULT_CORE_TRANSCRIBE_AUDIO = core_transcribe_audio


def _resolve_proxy_compat_callable(name: str, default: object) -> object:
    local_candidate = globals()[name]
    if local_candidate is not default:
        return local_candidate
    try:
        from app.modules.proxy import service as proxy_service_module
    except Exception:
        return local_candidate
    service_candidate = getattr(proxy_service_module, name, local_candidate)
    if service_candidate is not default:
        return service_candidate
    return local_candidate


@dataclass(frozen=True, slots=True)
class RequestCapabilities:
    route_family: str
    route_class: str
    transport: str
    model: str | None
    continuity_param: str | None = None


@dataclass(frozen=True, slots=True)
class ProviderSubject:
    provider_kind: ProviderKind
    routing_subject_id: str
    account: Account | None = None
    api_key_encrypted: bytes | None = None
    organization_id: str | None = None
    project_id: str | None = None

    def require_account(self) -> Account:
        if self.account is None:
            raise ValueError("Provider subject does not include a ChatGPT-web account")
        return self.account

    def require_api_key_encrypted(self) -> bytes:
        if self.api_key_encrypted is None:
            raise ValueError("Provider subject does not include encrypted OpenAI Platform credentials")
        return self.api_key_encrypted


@dataclass(frozen=True, slots=True)
class ProviderCapabilityDecision:
    allowed: bool
    error_code: str | None = None
    error_message: str | None = None
    error_param: str | None = None
    rejection_reason: str | None = None


@dataclass(frozen=True, slots=True)
class ProviderModelsResult:
    payload: dict[str, JsonValue]
    upstream_request_id: str | None


@dataclass(frozen=True, slots=True)
class ProviderCreateResponseResult:
    payload: OpenAIResponsePayload | dict[str, JsonValue]
    upstream_request_id: str | None


@dataclass(frozen=True, slots=True)
class ProviderStreamResponseResult:
    event_stream: AsyncIterator[str]
    upstream_request_id: str | None


class ProviderAdapter(Protocol):
    provider_kind: ProviderKind

    async def ensure_ready(
        self,
        subject: ProviderSubject,
        *,
        force: bool = False,
        timeout_seconds: float | None = None,
    ) -> ProviderSubject: ...

    def check_capabilities(
        self,
        subject: ProviderSubject,
        capabilities: RequestCapabilities,
    ) -> ProviderCapabilityDecision: ...

    async def fetch_models(self, subject: ProviderSubject) -> ProviderModelsResult: ...

    async def create_response(
        self,
        subject: ProviderSubject,
        payload: Mapping[str, JsonValue],
    ) -> ProviderCreateResponseResult: ...

    async def stream_responses(
        self,
        subject: ProviderSubject,
        payload: Mapping[str, JsonValue],
    ) -> ProviderStreamResponseResult: ...


class ChatGPTWebProviderAdapter:
    provider_kind: ProviderKind = CHATGPT_WEB_PROVIDER_KIND

    def __init__(self, repo_factory: ProxyRepoFactory) -> None:
        self._repo_factory = repo_factory
        self._encryptor = TokenEncryptor()

    def _upstream_auth(self, subject: ProviderSubject) -> tuple[str, str | None]:
        account = subject.require_account()
        access_token = self._encryptor.decrypt(account.access_token_encrypted)
        return access_token, _header_account_id(account.chatgpt_account_id)

    async def ensure_ready(
        self,
        subject: ProviderSubject,
        *,
        force: bool = False,
        timeout_seconds: float | None = None,
    ) -> ProviderSubject:
        account = subject.require_account()
        async with self._repo_factory() as repos:
            auth_manager = AuthManager(repos.accounts)
            refreshed = await auth_manager.ensure_fresh(account, force=force)
        return replace(subject, account=refreshed)

    def check_capabilities(
        self,
        subject: ProviderSubject,
        capabilities: RequestCapabilities,
    ) -> ProviderCapabilityDecision:
        del subject, capabilities
        return ProviderCapabilityDecision(allowed=True)

    async def fetch_models(self, subject: ProviderSubject) -> ProviderModelsResult:
        del subject
        raise NotImplementedError("ChatGPT-web model discovery remains local-registry backed in phase 1")

    async def create_response(
        self,
        subject: ProviderSubject,
        payload: Mapping[str, JsonValue],
    ) -> ProviderCreateResponseResult:
        del subject, payload
        raise NotImplementedError("ChatGPT-web response execution stays on existing proxy paths in phase 1")

    async def stream_responses(
        self,
        subject: ProviderSubject,
        payload: Mapping[str, JsonValue],
    ) -> ProviderStreamResponseResult:
        del subject, payload
        raise NotImplementedError("ChatGPT-web streaming stays on existing proxy paths in phase 1")

    async def compact_response(
        self,
        subject: ProviderSubject,
        payload: ResponsesCompactRequest,
        headers: Mapping[str, str],
    ) -> CompactResponsePayload:
        access_token, account_id = self._upstream_auth(subject)
        compact_impl = cast(
            _CompactResponsesCallable,
            _resolve_proxy_compat_callable("core_compact_responses", _DEFAULT_CORE_COMPACT_RESPONSES),
        )
        return await compact_impl(payload, headers, access_token, account_id)

    async def stream_response_events(
        self,
        subject: ProviderSubject,
        payload: ResponsesRequest,
        headers: Mapping[str, str],
        *,
        raise_for_status: bool = True,
        upstream_stream_transport: str | None = None,
    ) -> AsyncIterator[str]:
        access_token, account_id = self._upstream_auth(subject)
        stream_impl = cast(
            "Callable[..., AsyncIterator[str]]",
            _resolve_proxy_compat_callable("core_stream_responses", _DEFAULT_CORE_STREAM_RESPONSES),
        )
        kwargs: dict[str, object] = {"raise_for_status": raise_for_status}
        stream_signature = inspect.signature(stream_impl)
        supports_transport_override = any(
            parameter.kind is inspect.Parameter.VAR_KEYWORD or parameter.name == "upstream_stream_transport_override"
            for parameter in stream_signature.parameters.values()
        )
        if supports_transport_override:
            kwargs["upstream_stream_transport_override"] = upstream_stream_transport
        return stream_impl(payload, headers, access_token, account_id, **kwargs)

    async def transcribe_audio(
        self,
        subject: ProviderSubject,
        *,
        audio_bytes: bytes,
        filename: str,
        content_type: str | None,
        prompt: str | None,
        headers: Mapping[str, str],
    ) -> dict[str, JsonValue]:
        access_token, account_id = self._upstream_auth(subject)
        transcribe_impl = cast(
            _TranscribeAudioCallable,
            _resolve_proxy_compat_callable("core_transcribe_audio", _DEFAULT_CORE_TRANSCRIBE_AUDIO),
        )
        return await transcribe_impl(
            audio_bytes,
            filename=filename,
            content_type=content_type,
            prompt=prompt,
            headers=headers,
            access_token=access_token,
            account_id=account_id,
        )

    async def open_responses_websocket(
        self,
        subject: ProviderSubject,
        headers: Mapping[str, str],
    ) -> UpstreamResponsesWebSocket:
        access_token, account_id = self._upstream_auth(subject)
        return await connect_responses_websocket(dict(headers), access_token, account_id)

    async def refresh_usage(
        self,
        repos: ProxyRepositories,
        subjects: Sequence[ProviderSubject],
    ) -> None:
        accounts = [subject.account for subject in subjects if subject.account is not None]
        if not accounts:
            return
        latest_usage = await repos.usage.latest_by_account()
        updater = UsageUpdater(repos.usage, repos.accounts, repos.additional_usage)
        await updater.refresh_accounts(accounts, latest_usage)


class OpenAIPlatformProviderAdapter:
    provider_kind: ProviderKind = OPENAI_PLATFORM_PROVIDER_KIND

    def __init__(self, *, base_url: str = "https://api.openai.com") -> None:
        self._base_url = base_url
        self._encryptor = TokenEncryptor()

    def _maybe_log_request_start(
        self,
        *,
        kind: str,
        method: str,
        subject: ProviderSubject,
    ) -> float:
        started_at = time.monotonic()
        if not get_settings().log_upstream_request_summary:
            return started_at
        logger.info(
            (
                "upstream_request_start request_id=%s kind=%s method=%s target=%s "
                "provider_kind=%s route_class=%s routing_subject_id=%s account_id=%s upstream_request_id=%s"
            ),
            get_request_id(),
            kind,
            method,
            self._base_url,
            self.provider_kind,
            OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
            subject.routing_subject_id,
            None,
            None,
        )
        return started_at

    def _maybe_log_request_complete(
        self,
        *,
        kind: str,
        method: str,
        subject: ProviderSubject,
        started_at: float,
        status_code: int | None,
        error_code: str | None,
        error_message: str | None,
        upstream_request_id: str | None = None,
    ) -> None:
        if not get_settings().log_upstream_request_summary:
            return
        level = logging.INFO
        if status_code is not None and status_code >= 500:
            level = logging.ERROR
        elif (status_code is not None and status_code >= 400) or error_code is not None:
            level = logging.WARNING
        logger.log(
            level,
            (
                "upstream_request_complete request_id=%s kind=%s method=%s target=%s "
                "provider_kind=%s route_class=%s routing_subject_id=%s account_id=%s "
                "status=%s duration_ms=%s error_code=%s error_message=%s upstream_request_id=%s"
            ),
            get_request_id(),
            kind,
            method,
            self._base_url,
            self.provider_kind,
            OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
            subject.routing_subject_id,
            None,
            status_code,
            int((time.monotonic() - started_at) * 1000),
            error_code,
            error_message,
            upstream_request_id,
        )

    async def ensure_ready(
        self,
        subject: ProviderSubject,
        *,
        force: bool = False,
        timeout_seconds: float | None = None,
    ) -> ProviderSubject:
        del force
        del timeout_seconds
        return subject

    def check_capabilities(
        self,
        subject: ProviderSubject,
        capabilities: RequestCapabilities,
    ) -> ProviderCapabilityDecision:
        del subject
        if capabilities.transport == "websocket":
            return ProviderCapabilityDecision(
                allowed=False,
                error_code="provider_transport_unsupported",
                error_message="OpenAI Platform identities do not support websocket /v1/responses routes in phase 1.",
                error_param="transport",
                rejection_reason="provider_transport_unsupported",
            )
        if capabilities.continuity_param is not None:
            return ProviderCapabilityDecision(
                allowed=False,
                error_code="provider_continuity_unsupported",
                error_message=(
                    f"OpenAI Platform identities do not support "
                    f"'{capabilities.continuity_param}' continuity in phase 1."
                ),
                error_param=capabilities.continuity_param,
                rejection_reason="platform_continuity_unsupported",
            )
        if capabilities.route_family not in PHASE1_PLATFORM_ROUTE_FAMILIES:
            return ProviderCapabilityDecision(
                allowed=False,
                error_code="provider_feature_unsupported",
                error_message=f"OpenAI Platform identities do not support {capabilities.route_family} in phase 1.",
                rejection_reason="provider_feature_unsupported",
            )
        return ProviderCapabilityDecision(allowed=True)

    async def fetch_models(self, subject: ProviderSubject) -> ProviderModelsResult:
        started_at = self._maybe_log_request_start(kind="platform_models", method="GET", subject=subject)
        try:
            result = await fetch_platform_models(
                base_url=self._base_url,
                api_key=self._encryptor.decrypt(subject.require_api_key_encrypted()),
                organization=subject.organization_id,
                project=subject.project_id,
            )
        except OpenAIPlatformError as exc:
            self._maybe_log_request_complete(
                kind="platform_models",
                method="GET",
                subject=subject,
                started_at=started_at,
                status_code=exc.status_code,
                error_code=_platform_error_code(exc.payload),
                error_message=_platform_error_message(exc.payload),
            )
            raise
        self._maybe_log_request_complete(
            kind="platform_models",
            method="GET",
            subject=subject,
            started_at=started_at,
            status_code=200,
            error_code=None,
            error_message=None,
            upstream_request_id=result.upstream_request_id,
        )
        return ProviderModelsResult(payload=result.payload, upstream_request_id=result.upstream_request_id)

    async def create_response(
        self,
        subject: ProviderSubject,
        payload: Mapping[str, JsonValue],
    ) -> ProviderCreateResponseResult:
        started_at = self._maybe_log_request_start(kind="platform_responses", method="POST", subject=subject)
        try:
            result = await create_platform_response(
                base_url=self._base_url,
                payload=payload,
                api_key=self._encryptor.decrypt(subject.require_api_key_encrypted()),
                organization=subject.organization_id,
                project=subject.project_id,
            )
        except OpenAIPlatformError as exc:
            self._maybe_log_request_complete(
                kind="platform_responses",
                method="POST",
                subject=subject,
                started_at=started_at,
                status_code=exc.status_code,
                error_code=_platform_error_code(exc.payload),
                error_message=_platform_error_message(exc.payload),
            )
            raise
        status = _platform_response_status(result.payload)
        self._maybe_log_request_complete(
            kind="platform_responses",
            method="POST",
            subject=subject,
            started_at=started_at,
            status_code=200,
            error_code=None if status != "failed" else _platform_response_error_code(result.payload),
            error_message=None if status != "failed" else _platform_response_error_message(result.payload),
            upstream_request_id=result.upstream_request_id,
        )
        return ProviderCreateResponseResult(
            payload=result.payload,
            upstream_request_id=result.upstream_request_id,
        )

    async def stream_responses(
        self,
        subject: ProviderSubject,
        payload: Mapping[str, JsonValue],
    ) -> ProviderStreamResponseResult:
        started_at = self._maybe_log_request_start(kind="platform_responses", method="POST", subject=subject)
        try:
            result = await stream_platform_responses(
                base_url=self._base_url,
                payload=payload,
                api_key=self._encryptor.decrypt(subject.require_api_key_encrypted()),
                organization=subject.organization_id,
                project=subject.project_id,
            )
        except OpenAIPlatformError as exc:
            self._maybe_log_request_complete(
                kind="platform_responses",
                method="POST",
                subject=subject,
                started_at=started_at,
                status_code=exc.status_code,
                error_code=_platform_error_code(exc.payload),
                error_message=_platform_error_message(exc.payload),
            )
            raise
        self._maybe_log_request_complete(
            kind="platform_responses",
            method="POST",
            subject=subject,
            started_at=started_at,
            status_code=200,
            error_code=None,
            error_message=None,
            upstream_request_id=result.upstream_request_id,
        )
        return ProviderStreamResponseResult(
            event_stream=result.event_stream,
            upstream_request_id=result.upstream_request_id,
        )

    async def validate_identity(
        self,
        *,
        api_key: str,
        organization: str | None = None,
        project: str | None = None,
    ) -> ProviderModelsResult:
        subject = ProviderSubject(
            provider_kind=self.provider_kind,
            routing_subject_id="platform_validation",
        )
        started_at = self._maybe_log_request_start(kind="platform_models_validate", method="GET", subject=subject)
        try:
            result = await validate_platform_identity(
                base_url=self._base_url,
                api_key=api_key,
                organization=organization,
                project=project,
            )
        except OpenAIPlatformError as exc:
            self._maybe_log_request_complete(
                kind="platform_models_validate",
                method="GET",
                subject=subject,
                started_at=started_at,
                status_code=exc.status_code,
                error_code=_platform_error_code(exc.payload),
                error_message=_platform_error_message(exc.payload),
            )
            raise
        self._maybe_log_request_complete(
            kind="platform_models_validate",
            method="GET",
            subject=subject,
            started_at=started_at,
            status_code=200,
            error_code=None,
            error_message=None,
            upstream_request_id=result.upstream_request_id,
        )
        return ProviderModelsResult(payload=result.payload, upstream_request_id=result.upstream_request_id)


def _platform_error_code(payload: Mapping[str, JsonValue]) -> str | None:
    error = payload.get("error")
    if not is_json_mapping(error):
        return None
    code = error.get("code")
    return code if isinstance(code, str) else None


def _platform_error_message(payload: Mapping[str, JsonValue]) -> str | None:
    error = payload.get("error")
    if not is_json_mapping(error):
        return None
    message = error.get("message")
    return message if isinstance(message, str) else None


def _platform_response_status(payload: OpenAIResponsePayload | dict[str, JsonValue]) -> str | None:
    if isinstance(payload, OpenAIResponsePayload):
        return payload.status
    status = payload.get("status")
    return status if isinstance(status, str) else None


def _platform_response_error_code(payload: OpenAIResponsePayload | dict[str, JsonValue]) -> str | None:
    if isinstance(payload, OpenAIResponsePayload):
        return payload.error.code if payload.error is not None else None
    return _platform_error_code(payload)


def _platform_response_error_message(payload: OpenAIResponsePayload | dict[str, JsonValue]) -> str | None:
    if isinstance(payload, OpenAIResponsePayload):
        return payload.error.message if payload.error is not None else None
    return _platform_error_message(payload)



"""add dashboard_settings.routing_strategy

Revision ID: 20260225_000000_add_dashboard_settings_routing_strategy
Revises: 20260218_000100_add_import_without_overwrite_and_drop_accounts_email_unique
Create Date: 2026-02-25
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260225_000000_add_dashboard_settings_routing_strategy"
down_revision = "20260218_000100_add_import_without_overwrite_and_drop_accounts_email_unique"
branch_labels = None
depends_on = None


def _table_exists(connection: Connection, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    return inspector.has_table(table_name)


def _columns(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {str(column["name"]) for column in inspector.get_columns(table_name) if column.get("name") is not None}


def upgrade() -> None:
    bind = op.get_bind()
    if not _table_exists(bind, "dashboard_settings"):
        return
    columns = _columns(bind, "dashboard_settings")
    if "routing_strategy" in columns:
        return
    with op.batch_alter_table("dashboard_settings") as batch_op:
        batch_op.add_column(
            sa.Column(
                "routing_strategy",
                sa.String(),
                nullable=False,
                server_default=sa.text("'usage_weighted'"),
            )
        )


def downgrade() -> None:
    # Downgrade intentionally keeps column to avoid SQLite table rebuild risk.
    return



from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import cast

import pytest
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse, Response
from starlette.types import Message

from app.core.middleware.request_id import add_request_id_middleware
from app.core.utils.request_id import get_request_id

pytestmark = pytest.mark.unit

_Dispatch = Callable[[Request, Callable[[Request], Awaitable[Response]]], Awaitable[Response]]


@pytest.mark.asyncio
async def test_request_id_middleware_resets_context_on_success():
    app = FastAPI()
    add_request_id_middleware(app)
    dispatch = cast(_Dispatch, app.user_middleware[0].kwargs["dispatch"])

    request = Request(
        {
            "type": "http",
            "http_version": "1.1",
            "method": "GET",
            "scheme": "http",
            "path": "/health",
            "raw_path": b"/health",
            "query_string": b"",
            "root_path": "",
            "headers": [(b"x-request-id", b"req-test-123")],
            "client": ("testclient", 50000),
            "server": ("testserver", 80),
        },
        receive=_empty_receive,
    )

    async def call_next(_: Request) -> JSONResponse:
        assert get_request_id() == "req-test-123"
        return JSONResponse({"ok": True})

    response = await dispatch(request, call_next)

    assert response.headers["x-request-id"] == "req-test-123"
    assert get_request_id() is None


async def _empty_receive() -> Message:
    return {"type": "http.request", "body": b"", "more_body": False}



from __future__ import annotations

from datetime import datetime, timedelta

from app.core import usage as usage_core
from app.core.crypto import TokenEncryptor
from app.core.usage.types import UsageWindowRow
from app.core.utils.time import utcnow
from app.db.models import UsageHistory
from app.modules.accounts.mappers import build_account_summaries
from app.modules.dashboard.builders import (
    build_dashboard_overview_summary,
    build_overview_timeframe,
    resolve_overview_timeframe,
)
from app.modules.dashboard.repository import DashboardRepository
from app.modules.dashboard.schemas import (
    DashboardOverviewResponse,
    DashboardOverviewTimeframeKey,
    DashboardUsageWindows,
    DepletionResponse,
)
from app.modules.usage.builders import (
    align_bucket_window_start,
    build_activity_summaries,
    build_trends_from_buckets,
    build_usage_window_response,
)
from app.modules.usage.depletion_service import (
    compute_aggregate_depletion,
    compute_depletion_for_account,
)


class DashboardService:
    def __init__(self, repo: DashboardRepository) -> None:
        self._repo = repo
        self._encryptor = TokenEncryptor()

    async def get_overview(
        self,
        timeframe_key: DashboardOverviewTimeframeKey = "7d",
    ) -> DashboardOverviewResponse:
        now = utcnow()
        overview_timeframe = resolve_overview_timeframe(timeframe_key)
        accounts = await self._repo.list_accounts()
        primary_usage = await self._repo.latest_usage_by_account("primary")
        secondary_usage = await self._repo.latest_usage_by_account("secondary")

        account_summaries = build_account_summaries(
            accounts=accounts,
            primary_usage=primary_usage,
            secondary_usage=secondary_usage,
            encryptor=self._encryptor,
            include_auth=False,
        )

        primary_rows_raw = _rows_from_latest(primary_usage)
        secondary_rows_raw = _rows_from_latest(secondary_usage)
        primary_rows, secondary_rows = usage_core.normalize_weekly_only_rows(
            primary_rows_raw,
            secondary_rows_raw,
        )

        bucket_since = now - timedelta(minutes=overview_timeframe.window_minutes)
        bucket_query_since = align_bucket_window_start(
            bucket_since,
            overview_timeframe.bucket_seconds,
        )
        bucket_rows = await self._repo.aggregate_logs_by_bucket(
            bucket_query_since,
            overview_timeframe.bucket_seconds,
        )
        trends, _, _ = build_trends_from_buckets(
            bucket_rows,
            bucket_since,
            bucket_seconds=overview_timeframe.bucket_seconds,
            bucket_count=overview_timeframe.bucket_count,
        )
        activity_aggregate = await self._repo.aggregate_activity_since(bucket_since)
        top_error = await self._repo.top_error_since(bucket_since)
        activity_metrics, activity_cost = build_activity_summaries(
            activity_aggregate,
            top_error=top_error,
        )

        summary = build_dashboard_overview_summary(
            accounts=accounts,
            primary_rows=primary_rows,
            secondary_rows=secondary_rows,
            activity_metrics=activity_metrics,
            activity_cost=activity_cost,
        )

        secondary_minutes = usage_core.resolve_window_minutes("secondary", secondary_rows)
        primary_window_minutes = usage_core.resolve_window_minutes("primary", primary_rows)

        windows = DashboardUsageWindows(
            primary=build_usage_window_response(
                window_key="primary",
                window_minutes=primary_window_minutes,
                usage_rows=primary_rows,
                accounts=accounts,
            ),
            secondary=build_usage_window_response(
                window_key="secondary",
                window_minutes=secondary_minutes,
                usage_rows=secondary_rows,
                accounts=accounts,
            ),
        )

        # Compute depletion separately for primary-window and secondary-window
        # accounts so the aggregate is not skewed by mixing different window
        # durations.  The response includes a "window" field that tells the
        # frontend which donut to render the safe-line marker on.
        normalized_primary_ids = {row.account_id for row in primary_rows}
        all_account_ids = set(primary_usage.keys()) | set(secondary_usage.keys())

        # Batch fetch: collect account IDs and determine the widest lookback
        # per window so we can issue at most 2 bulk queries instead of O(N).
        pri_fetch_ids: list[str] = []
        sec_fetch_ids: list[str] = []
        pri_since = now  # will be narrowed to the earliest needed
        sec_since = now
        # Per-account cutoffs for in-memory filtering after bulk fetch
        pri_cutoffs: dict[str, datetime] = {}
        sec_cutoffs: dict[str, datetime] = {}
        weekly_only_ids: set[str] = set()
        weekly_only_history_sources: dict[str, str] = {}

        for account_id in all_account_ids:
            if account_id in normalized_primary_ids:
                usage_entry = primary_usage[account_id]
                acct_window = usage_entry.window_minutes if usage_entry.window_minutes else 300
                acct_since = now - timedelta(minutes=acct_window)
                pri_fetch_ids.append(account_id)
                pri_cutoffs[account_id] = acct_since
                if acct_since < pri_since:
                    pri_since = acct_since
                if account_id in secondary_usage:
                    sec_entry = secondary_usage[account_id]
                    sec_window = sec_entry.window_minutes if sec_entry.window_minutes else 10080
                    s_since = now - timedelta(minutes=sec_window)
                    sec_fetch_ids.append(account_id)
                    sec_cutoffs[account_id] = s_since
                    if s_since < sec_since:
                        sec_since = s_since
            elif account_id in primary_usage:
                weekly_only_ids.add(account_id)
                primary_entry = primary_usage[account_id]
                sec_entry = secondary_usage.get(account_id)
                use_primary_stream = _should_use_weekly_primary_history(primary_entry, sec_entry)
                weekly_only_history_sources[account_id] = "primary" if use_primary_stream else "secondary"
                current_entry = primary_entry if use_primary_stream else sec_entry
                acct_window = current_entry.window_minutes if current_entry and current_entry.window_minutes else 10080
                acct_since = now - timedelta(minutes=acct_window)
                if use_primary_stream:
                    pri_fetch_ids.append(account_id)
                    pri_cutoffs[account_id] = acct_since
                    if acct_since < pri_since:
                        pri_since = acct_since
                else:
                    sec_fetch_ids.append(account_id)
                    sec_cutoffs[account_id] = acct_since
                    if acct_since < sec_since:
                        sec_since = acct_since
            else:
                sec_entry = secondary_usage[account_id]
                acct_window = sec_entry.window_minutes if sec_entry.window_minutes else 10080
                acct_since = now - timedelta(minutes=acct_window)
                sec_fetch_ids.append(account_id)
                sec_cutoffs[account_id] = acct_since
                if acct_since < sec_since:
                    sec_since = acct_since

        # Issue at most 2 bulk queries
        all_pri_rows = (
            await self._repo.bulk_usage_history_since(pri_fetch_ids, "primary", pri_since) if pri_fetch_ids else {}
        )
        all_sec_rows = (
            await self._repo.bulk_usage_history_since(sec_fetch_ids, "secondary", sec_since) if sec_fetch_ids else {}
        )

        # Filter in-memory to each account's actual cutoff
        primary_history: dict[str, list[UsageHistory]] = {}
        secondary_history: dict[str, list[UsageHistory]] = {}

        for account_id in all_account_ids:
            if account_id in normalized_primary_ids:
                cutoff = pri_cutoffs[account_id]
                rows = [r for r in all_pri_rows.get(account_id, []) if r.recorded_at >= cutoff]
                if rows:
                    primary_history[account_id] = rows
                if account_id in sec_cutoffs:
                    s_cutoff = sec_cutoffs[account_id]
                    s_rows = [r for r in all_sec_rows.get(account_id, []) if r.recorded_at >= s_cutoff]
                    if s_rows:
                        secondary_history[account_id] = s_rows
            elif account_id in weekly_only_ids:
                source = weekly_only_history_sources[account_id]
                if source == "primary":
                    cutoff = pri_cutoffs[account_id]
                    rows = [r for r in all_pri_rows.get(account_id, []) if r.recorded_at >= cutoff]
                else:
                    cutoff = sec_cutoffs[account_id]
                    rows = [r for r in all_sec_rows.get(account_id, []) if r.recorded_at >= cutoff]
                if rows:
                    secondary_history[account_id] = rows
            else:
                cutoff = sec_cutoffs[account_id]
                rows = [r for r in all_sec_rows.get(account_id, []) if r.recorded_at >= cutoff]
                if rows:
                    secondary_history[account_id] = rows

        pri_depletion, sec_depletion = _build_depletion_by_window(primary_history, secondary_history, now)

        additional_ts = await self._repo.latest_additional_recorded_at()
        return DashboardOverviewResponse(
            last_sync_at=_latest_recorded_at(primary_usage, secondary_usage, additional_ts),
            timeframe=build_overview_timeframe(overview_timeframe),
            accounts=account_summaries,
            summary=summary,
            windows=windows,
            trends=trends,
            depletion_primary=pri_depletion,
            depletion_secondary=sec_depletion,
        )


def _build_depletion_by_window(
    primary_history: dict[str, list[UsageHistory]],
    secondary_history: dict[str, list[UsageHistory]],
    now,
) -> tuple[DepletionResponse | None, DepletionResponse | None]:
    """Compute depletion independently per window."""

    def _aggregate(history: dict[str, list[UsageHistory]], window: str) -> DepletionResponse | None:
        metrics = []
        for account_id, rows in history.items():
            m = compute_depletion_for_account(
                account_id=account_id,
                limit_name="standard",
                window=window,
                history=rows,
                now=now,
            )
            metrics.append(m)
        agg = compute_aggregate_depletion(metrics)
        if agg is None:
            return None
        return DepletionResponse(
            risk=agg.risk,
            risk_level=agg.risk_level,
            burn_rate=agg.burn_rate,
            safe_usage_percent=agg.safe_usage_percent,
            projected_exhaustion_at=agg.projected_exhaustion_at,
            seconds_until_exhaustion=agg.seconds_until_exhaustion,
        )

    return _aggregate(primary_history, "primary"), _aggregate(secondary_history, "secondary")


def _rows_from_latest(latest: dict[str, UsageHistory]) -> list[UsageWindowRow]:
    return [
        UsageWindowRow(
            account_id=entry.account_id,
            used_percent=entry.used_percent,
            reset_at=entry.reset_at,
            window_minutes=entry.window_minutes,
            recorded_at=entry.recorded_at,
        )
        for entry in latest.values()
    ]


def _should_use_weekly_primary_history(
    primary_entry: UsageHistory,
    secondary_entry: UsageHistory | None,
) -> bool:
    return usage_core.should_use_weekly_primary(
        _usage_history_to_window_row(primary_entry),
        _usage_history_to_window_row(secondary_entry) if secondary_entry is not None else None,
    )


def _usage_history_to_window_row(entry: UsageHistory) -> UsageWindowRow:
    return UsageWindowRow(
        account_id=entry.account_id,
        used_percent=entry.used_percent,
        reset_at=entry.reset_at,
        window_minutes=entry.window_minutes,
        recorded_at=entry.recorded_at,
    )


def _latest_recorded_at(
    primary_usage: dict[str, UsageHistory],
    secondary_usage: dict[str, UsageHistory],
    additional_ts: datetime | None = None,
):
    timestamps = [
        entry.recorded_at
        for entry in list(primary_usage.values()) + list(secondary_usage.values())
        if entry.recorded_at is not None
    ]
    if additional_ts is not None:
        timestamps.append(additional_ts)
    return max(timestamps) if timestamps else None



from __future__ import annotations

import sqlalchemy as sa
from alembic import op

revision = "20260328_140000_add_scheduler_leader_table"
down_revision = "20260328_130000_add_audit_logs_table"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    inspector = sa.inspect(bind)
    if not inspector.has_table("scheduler_leader"):
        op.create_table(
            "scheduler_leader",
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("leader_id", sa.String(length=100), nullable=False),
            sa.Column("acquired_at", sa.DateTime(timezone=True), nullable=False),
            sa.Column("expires_at", sa.DateTime(timezone=True), nullable=False),
            sa.PrimaryKeyConstraint("id"),
        )

    index_names = {index["name"] for index in inspector.get_indexes("scheduler_leader")}
    if "ix_scheduler_leader_expires_at" not in index_names:
        op.create_index("ix_scheduler_leader_expires_at", "scheduler_leader", ["expires_at"], unique=False)


def downgrade() -> None:
    bind = op.get_bind()
    inspector = sa.inspect(bind)
    if not inspector.has_table("scheduler_leader"):
        return

    index_names = {index["name"] for index in inspector.get_indexes("scheduler_leader")}
    if "ix_scheduler_leader_expires_at" in index_names:
        op.drop_index("ix_scheduler_leader_expires_at", table_name="scheduler_leader")
    op.drop_table("scheduler_leader")



from __future__ import annotations

from datetime import timedelta

import pytest

from app.core.crypto import TokenEncryptor
from app.core.utils.time import utcnow
from app.db.models import Account, AccountStatus
from app.db.session import SessionLocal
from app.modules.accounts.repository import AccountsRepository
from app.modules.request_logs.repository import RequestLogsRepository

pytestmark = pytest.mark.integration


def _make_account(account_id: str, email: str) -> Account:
    encryptor = TokenEncryptor()
    return Account(
        id=account_id,
        email=email,
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access"),
        refresh_token_encrypted=encryptor.encrypt("refresh"),
        id_token_encrypted=encryptor.encrypt("id"),
        last_refresh=utcnow(),
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )


@pytest.mark.asyncio
async def test_request_logs_options_returns_distinct_accounts_and_models(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_opt_a", "a@example.com"))
        await accounts_repo.upsert(_make_account("acc_opt_b", "b@example.com"))

        await logs_repo.add_log(
            account_id="acc_opt_a",
            request_id="req_opt_1",
            model="gpt-5.1",
            input_tokens=10,
            output_tokens=10,
            latency_ms=100,
            status="success",
            error_code=None,
            requested_at=now - timedelta(minutes=1),
        )
        await logs_repo.add_log(
            account_id="acc_opt_b",
            request_id="req_opt_2",
            model="gpt-4o",
            input_tokens=10,
            output_tokens=10,
            latency_ms=100,
            status="error",
            error_code="rate_limit_exceeded",
            error_message="Rate limit reached",
            requested_at=now,
        )

    response = await async_client.get("/api/request-logs/options")
    assert response.status_code == 200
    payload = response.json()
    assert payload["accountIds"] == ["acc_opt_a", "acc_opt_b"]
    assert payload["modelOptions"] == [
        {"model": "gpt-4o", "reasoningEffort": None},
        {"model": "gpt-5.1", "reasoningEffort": None},
    ]
    assert payload["statuses"] == ["ok", "rate_limit"]


@pytest.mark.asyncio
async def test_request_logs_options_ignores_status_self_filter(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_opt_ok", "ok@example.com"))
        await accounts_repo.upsert(_make_account("acc_opt_err", "err@example.com"))

        await logs_repo.add_log(
            account_id="acc_opt_ok",
            request_id="req_opt_ok",
            model="gpt-5.1",
            input_tokens=10,
            output_tokens=10,
            latency_ms=100,
            status="success",
            error_code=None,
            requested_at=now,
        )
        await logs_repo.add_log(
            account_id="acc_opt_err",
            request_id="req_opt_err",
            model="gpt-4o",
            input_tokens=10,
            output_tokens=10,
            latency_ms=100,
            status="error",
            error_code="rate_limit_exceeded",
            requested_at=now,
        )

    response = await async_client.get("/api/request-logs/options?status=ok")
    assert response.status_code == 200
    payload = response.json()
    assert payload["accountIds"] == ["acc_opt_err", "acc_opt_ok"]
    assert payload["modelOptions"] == [
        {"model": "gpt-4o", "reasoningEffort": None},
        {"model": "gpt-5.1", "reasoningEffort": None},
    ]
    assert payload["statuses"] == ["ok", "rate_limit"]


@pytest.mark.asyncio
async def test_request_logs_options_ignore_status_matches_unfiltered_response(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_opt_ok_2", "ok2@example.com"))
        await accounts_repo.upsert(_make_account("acc_opt_quota", "quota@example.com"))

        await logs_repo.add_log(
            account_id="acc_opt_ok_2",
            request_id="req_opt_ok_2",
            model="gpt-5.1",
            input_tokens=10,
            output_tokens=10,
            latency_ms=100,
            status="success",
            error_code=None,
            requested_at=now,
        )
        await logs_repo.add_log(
            account_id="acc_opt_quota",
            request_id="req_opt_quota",
            model="gpt-5.2",
            input_tokens=10,
            output_tokens=10,
            latency_ms=100,
            status="error",
            error_code="insufficient_quota",
            requested_at=now,
        )

    base = await async_client.get("/api/request-logs/options")
    with_status = await async_client.get("/api/request-logs/options?status=ok&status=quota")

    assert base.status_code == 200
    assert with_status.status_code == 200
    assert with_status.json() == base.json()


@pytest.mark.asyncio
async def test_request_logs_options_respects_non_status_filters(async_client, db_setup):
    now = utcnow()
    old = now - timedelta(days=10)
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_scope_a", "scope-a@example.com"))
        await accounts_repo.upsert(_make_account("acc_scope_b", "scope-b@example.com"))

        await logs_repo.add_log(
            account_id="acc_scope_a",
            request_id="req_scope_1",
            model="gpt-5.1",
            input_tokens=10,
            output_tokens=10,
            latency_ms=100,
            status="success",
            error_code=None,
            requested_at=now,
        )
        await logs_repo.add_log(
            account_id="acc_scope_a",
            request_id="req_scope_2",
            model="gpt-5.1",
            input_tokens=10,
            output_tokens=10,
            latency_ms=100,
            status="error",
            error_code="rate_limit_exceeded",
            requested_at=now,
        )
        await logs_repo.add_log(
            account_id="acc_scope_b",
            request_id="req_scope_3",
            model="gpt-5.2",
            input_tokens=10,
            output_tokens=10,
            latency_ms=100,
            status="error",
            error_code="insufficient_quota",
            requested_at=old,
        )

    scoped = await async_client.get(
        "/api/request-logs/options"
        "?accountId=acc_scope_a"
        "&modelOption=gpt-5.1:::"
        f"&since={(now - timedelta(hours=1)).isoformat()}"
    )

    assert scoped.status_code == 200
    payload = scoped.json()
    assert payload["accountIds"] == ["acc_scope_a"]
    assert payload["modelOptions"] == [{"model": "gpt-5.1", "reasoningEffort": None}]
    assert payload["statuses"] == ["ok", "rate_limit"]
